#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

import functools
from typing import TYPE_CHECKING

from qt.core import (
    QAction,
    QApplication,
    QDialog,
    QEvent,
    QIcon,
    QLabel,
    QMenu,
    QModelIndex,
    QPixmap,
    QStackedWidget,
    QStatusBar,
    QStyle,
    QStyleOption,
    QStylePainter,
    Qt,
    QTabBar,
    QTimer,
    QToolButton,
    QUrl,
)

from calibre.constants import get_appname_for_display, get_version, ismacos
from calibre.customize.ui import find_plugin
from calibre.gui2 import config, error_dialog, gprefs, open_local_file, open_url, qapplication_or_fail
from calibre.gui2.book_details import BookDetails
from calibre.gui2.central import CentralContainer, LayoutButton
from calibre.gui2.layout_menu import LayoutMenu
from calibre.gui2.library.alternate_views import GridView
from calibre.gui2.library.bookshelf_view import BookshelfView
from calibre.gui2.library.views import BooksView, DeviceBooksView
from calibre.gui2.notify import get_notifier
from calibre.gui2.tag_browser.ui import TagBrowserWidget
from calibre.utils.config import prefs
from calibre.utils.icu import sort_key
from calibre.utils.localization import _, localize_website_link, ngettext

if TYPE_CHECKING:
    from calibre.gui2.ui import Main

_keep_refs = []


def partial(*args, **kwargs):
    ans = functools.partial(*args, **kwargs)
    _keep_refs.append(ans)
    return ans


class LibraryViewMixin:  # {{{
    library_view: BooksView

    def init_library_view_mixin(self: Main, db):
        self.library_view.files_dropped.connect(self.iactions['Add Books'].files_dropped, type=Qt.ConnectionType.QueuedConnection)
        self.library_view.books_dropped.connect(self.iactions['Edit Metadata'].books_dropped, type=Qt.ConnectionType.QueuedConnection)
        self.library_view.add_column_signal.connect(
            partial(
                self.iactions['Preferences'].do_config,
                initial_plugin=('Interface', 'Custom Columns'),
                close_after_initial=True,
            ),
            type=Qt.ConnectionType.QueuedConnection,
        )
        for func, args in [
            ('connect_to_search_box', (self.search, self.search_done)),
            ('connect_to_book_display', (self.book_details.show_data,)),
        ]:
            for view in (self.library_view, self.memory_view, self.card_a_view, self.card_b_view):
                getattr(view, func)(*args)

        self.memory_view.connect_dirtied_signal(self.upload_dirtied_booklists)
        self.memory_view.connect_upload_collections_signal(func=self.upload_collections, oncard=None)
        self.card_a_view.connect_dirtied_signal(self.upload_dirtied_booklists)
        self.card_a_view.connect_upload_collections_signal(func=self.upload_collections, oncard='carda')
        self.card_b_view.connect_dirtied_signal(self.upload_dirtied_booklists)
        self.card_b_view.connect_upload_collections_signal(func=self.upload_collections, oncard='cardb')
        self.book_on_device(None, reset=True)
        db.set_book_on_device_func(self.book_on_device)
        self.library_view.set_database(db)
        self.library_view._model.set_book_on_device_func(self.book_on_device)
        prefs['library_path'] = self.library_path

        for view in ('library', 'memory', 'card_a', 'card_b'):
            view = getattr(self, view + '_view')
            view.verticalHeader().sectionDoubleClicked.connect(self.iactions['View'].view_specific_book)

        self.library_view._model.set_highlight_only(config['highlight_search_matches'])

    def context_menu_action_hovered(self, ac):
        ac.showStatusText(self)

    def build_context_menus(self: Main):
        from calibre.gui2.bars import populate_menu

        def connect_hovered(menu):
            menu.hovered.connect(self.context_menu_action_hovered)
            for action in menu.actions():
                if m := action.menu():
                    connect_hovered(m)

        lm = QMenu(self)
        populate_menu(lm, gprefs['action-layout-context-menu'], self.iactions)
        connect_hovered(lm)
        dm = QMenu(self)
        populate_menu(dm, gprefs['action-layout-context-menu-device'], self.iactions)
        connect_hovered(dm)
        ec = self.iactions['Edit Collections'].qaction
        self.library_view.set_context_menu(lm, ec)
        sm = QMenu(self)
        populate_menu(sm, gprefs['action-layout-context-menu-split'], self.iactions)
        connect_hovered(sm)
        self.library_view.pin_view.set_context_menu(sm)
        for v in (self.memory_view, self.card_a_view, self.card_b_view):
            v.set_context_menu(dm, ec)

        if hasattr(self.cover_flow, 'set_context_menu'):
            cm = QMenu(self.cover_flow)
            populate_menu(cm, gprefs['action-layout-context-menu-cover-browser'], self.iactions)
            connect_hovered(cm)
            self.cover_flow.set_context_menu(cm)

    def search_done(self: Main, view, ok):
        if view is self.current_view():
            self.search.search_done(ok)
            self.set_number_of_books_shown()
            if ok:
                v = self.current_view()
                if hasattr(v, 'set_current_row'):
                    v.set_current_row(0)
                    if v is self.library_view and v.row_count() == 0:
                        self.book_details.reset_info()


# }}}


class UpdateLabel(QLabel):  # {{{
    def __init__(self, *args, **kwargs):
        QLabel.__init__(self, *args, **kwargs)
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def contextMenuEvent(self, ev):
        pass


# }}}


class VersionLabel(QLabel):  # {{{
    def __init__(self, parent):
        QLabel.__init__(self, parent)
        self.mouse_over = False
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setToolTip(_("See what's new in this calibre release"))

    def mouseReleaseEvent(self, ev):
        open_url(localize_website_link('https://calibre-ebook.com/whats-new'))
        ev.accept()
        return QLabel.mouseReleaseEvent(self, ev)

    def event(self, e):
        m = None
        et = e.type()
        if et == QEvent.Type.Enter:
            m = True
        elif et == QEvent.Type.Leave:
            m = False
        if m is not None and m != self.mouse_over:
            self.mouse_over = m
            self.update()
        return QLabel.event(self, e)

    def paintEvent(self, a0):
        if self.mouse_over:
            p = QStylePainter(self)
            tool = QStyleOption()
            tool.initFrom(self)
            tool.rect = self.rect()
            tool.state = QStyle.StateFlag.State_Raised | QStyle.StateFlag.State_Active | QStyle.StateFlag.State_MouseOver
            p.drawPrimitive(QStyle.PrimitiveElement.PE_PanelButtonTool, tool)
            p.end()
        return QLabel.paintEvent(self, a0)


# }}}


class StatusBar(QStatusBar):  # {{{
    def __init__(self, parent=None):
        QStatusBar.__init__(self, parent)
        self.version = get_version()
        self.base_msg = f'{get_appname_for_display()} {self.version}'
        self.device_string = ''
        self.update_label = UpdateLabel('')
        self.total = self.current = self.selected = self.library_total = 0
        self.addPermanentWidget(self.update_label)
        self.update_label.setVisible(False)
        self.defmsg = VersionLabel(self)
        self.addWidget(self.defmsg)
        self.set_label()

    def initialize(self, systray=None):
        self.systray = systray
        self.notifier = get_notifier(systray)

    def device_connected(self, devname):
        self.device_string = _('Connected ') + devname
        self.set_label()

    def update_state(self, library_total, total, current, selected):
        self.library_total = library_total
        self.total, self.current, self.selected = total, current, selected
        self.set_label()

    def set_label(self):
        try:
            self._set_label()
        except Exception:
            import traceback

            traceback.print_exc()

    def _set_label(self):
        msg = self.base_msg
        if self.device_string:
            msg += ' ..::.. ' + self.device_string
        else:
            msg += _(' %(created)s %(name)s') % dict(created=_('created by'), name='Kovid Goyal')

        if self.total != self.current:
            base = _('%(num)d of %(total)d books') % dict(num=self.current, total=self.total)
        else:
            base = ngettext('one book', '{} books', self.total).format(self.total)
        if self.selected > 0:
            base = ngettext('%(num)s, %(sel)d selected', '%(num)s, %(sel)d selected', self.selected) % dict(num=base, sel=self.selected)
        if self.library_total != self.total:
            base = _('{0}, {1} total').format(base, self.library_total)

        self.defmsg.setText(f'\xa0{msg}\xa0\xa0\xa0\xa0[{base}] ')
        self.clearMessage()

    def device_disconnected(self):
        self.device_string = ''
        self.set_label()

    def show_message(self, msg, timeout=0, show_notification=True):
        self.showMessage(msg, timeout)
        if self.notifier is not None and not config['disable_tray_notification'] and show_notification:
            self.notifier(msg)

    def clear_message(self):
        self.clearMessage()


# }}}


class AlternateViewsButtons(LayoutButton):  # {{{
    buttons = set()
    ignore_toggles = False
    needs_group_by = False

    def __init__(self, name: str, icon: str, label: str, view_name: str, gui: Main, shortcut=None, config_key=None):
        super().__init__(name, icon, label, gui.layout_container, shortcut=shortcut)
        self.set_state_to_show()
        self.action_toggle = QAction(self.icon(), _('Toggle') + ' ' + self.label, self)
        self.gui = gui
        self.gui.addAction(self.action_toggle)
        self.ck = config_key or name
        self.view_name = view_name
        if shortcut:
            self.gui.keyboard.register_shortcut(
                f'{self.ck} toggle {self.label}',
                str(self.action_toggle.text()),
                default_keys=(shortcut,),
                action=self.action_toggle,
                group=_('Main window layout'),
            )
        self.action_toggle.triggered.connect(self.toggle)
        self.action_toggle.changed.connect(self.update_shortcut)
        self.toggled.connect(self.update_state)
        self.toggled.connect(self.toggle_view)
        self.buttons.add(self)

    @property
    def is_visible(self):
        return self.isChecked()

    def update_state(self, *args):
        checked = args[0] if args else None
        if checked:
            self.set_state_to_hide()
        else:
            self.set_state_to_show()

    def save_state(self):
        gprefs[f'{self.ck} visible'] = bool(self.isChecked())

    def restore_state(self):
        if gprefs.get(f'{self.ck} visible', False):
            self.toggle()

    def toggle_view(self, show):
        if AlternateViewsButtons.ignore_toggles:
            return
        AlternateViewsButtons.ignore_toggles = True
        for btn in self.buttons:
            if btn == self:
                continue
            if btn.isChecked():
                btn.update_state(False)
        self.gui.library_view.alternate_views.show_view(self.view_name if show else None)
        self.gui.show_sort_button_for_alternate_view(show)
        self.gui.group_by_button.setVisible(self.needs_group_by and show)
        AlternateViewsButtons.ignore_toggles = False


# }}}


class GridViewButton(AlternateViewsButtons):  # {{{
    def __init__(self, gui):
        super().__init__(
            'cover_grid',
            'grid.png',
            _('Cover grid'),
            'grid',
            gui,
            shortcut='Alt+Shift+G',
            config_key='grid view',
        )


# }}}


class BookshelfViewButton(AlternateViewsButtons):  # {{{
    needs_group_by = True

    def __init__(self, gui):
        super().__init__(
            'bookshelf_view',
            'bookshelf.png',
            _('Bookshelf view'),
            'bookshelf',
            gui,
            shortcut='Alt+Shift+H',
            config_key='bookshelf view',
        )


# }}}


class SearchBarButton(LayoutButton):  # {{{
    def __init__(self, gui):
        sc = 'Alt+Shift+F'
        LayoutButton.__init__(self, 'search', 'search.png', _('Search bar'), gui, shortcut=sc)
        self.set_state_to_hide()
        self.action_toggle = QAction(self.icon(), _('Toggle') + ' ' + self.label, self)
        gui.addAction(self.action_toggle)
        gui.keyboard.register_shortcut(
            'search bar toggle' + self.label,
            str(self.action_toggle.text()),
            default_keys=(sc,),
            action=self.action_toggle,
            group=_('Main window layout'),
        )
        self.action_toggle.triggered.connect(self.toggle)
        self.action_toggle.changed.connect(self.update_shortcut)
        self.toggled.connect(self.update_state)
        self.update_shortcut()

    @property
    def is_visible(self):
        return self.isChecked()

    def update_state(self, *args):
        checked = args[0] if args else None
        if checked:
            self.set_state_to_hide()
        else:
            self.set_state_to_show()

    def save_state(self):
        gprefs['search bar visible'] = bool(self.isChecked())

    def restore_state(self):
        self.setChecked(bool(gprefs.get('search bar visible', True)))


# }}}


class VLTabs(QTabBar):  # {{{
    def __init__(self, parent):
        QTabBar.__init__(self, parent)
        self.setDocumentMode(True)
        self.setDrawBase(False)
        self.setTabsClosable(gprefs['vl_tabs_closable'])
        self.setMovable(self.tabsClosable())
        self.gui = parent
        self.ignore_tab_changed = False
        self.currentChanged.connect(self.tab_changed)
        self.tabMoved.connect(self.tab_moved, type=Qt.ConnectionType.QueuedConnection)
        self.tabCloseRequested.connect(self.tab_close)
        self.update_visibility()
        self.next_action = a = QAction(self)
        a.triggered.connect(partial(self.next_tab, delta=1)), self.gui.addAction(a)
        self.previous_action = a = QAction(self)
        a.triggered.connect(partial(self.next_tab, delta=-1)), self.gui.addAction(a)
        self.gui.keyboard.register_shortcut(
            'virtual-library-tab-bar-next',
            _('Next Virtual library'),
            action=self.next_action,
            default_keys=('Ctrl+Tab',),
            group=_('Virtual library'),
            description=_('Switch to the next Virtual library in the Virtual library tab bar'),
        )
        self.gui.keyboard.register_shortcut(
            'virtual-library-tab-bar-previous',
            _('Previous Virtual library'),
            action=self.previous_action,
            default_keys=('Ctrl+Shift+Tab',),
            group=_('Virtual library'),
            description=_('Switch to the previous Virtual library in the Virtual library tab bar'),
        )

    def next_tab(self, delta=1):
        if self.count() > 1 and self.isVisible():
            idx = (self.currentIndex() + delta) % self.count()
            self.setCurrentIndex(idx)

    def update_visibility(self):
        cv = self.gui.current_view()
        self.setVisible(gprefs['show_vl_tabs'] and self.count() > 1 and (cv is self.gui.library_view or cv is None))

    def enable_bar(self):
        gprefs['show_vl_tabs'] = True
        self.update_visibility()
        self.gui.set_number_of_books_shown()

    def disable_bar(self):
        gprefs['show_vl_tabs'] = False
        self.update_visibility()
        self.gui.set_number_of_books_shown()

    def lock_tab(self):
        gprefs['vl_tabs_closable'] = False
        self.setTabsClosable(False)
        self.setMovable(False)
        # Workaround for Qt bug where it doesn't recalculate the tab size after locking
        for idx in range(self.count()):
            self.setTabButton(idx, QTabBar.ButtonPosition.RightSide, None)
            self.setTabButton(idx, QTabBar.ButtonPosition.LeftSide, None)

    def unlock_tab(self):
        gprefs['vl_tabs_closable'] = True
        self.setTabsClosable(True)
        self.setMovable(True)
        # ensure no button on the All books tab since it is not closeable
        for idx in range(self.count()):
            if not self.tabData(idx):
                btn = self.tabButton(idx, QTabBar.ButtonPosition.RightSide)
                if btn is not None:
                    btn.setVisible(False)
                else:
                    btn = self.tabButton(idx, QTabBar.ButtonPosition.LeftSide)
                    if btn is not None:
                        btn.setVisible(False)
                break

    def tab_changed(self, idx):
        if self.ignore_tab_changed:
            return
        vl = str(self.tabData(idx) or '').strip() or None
        self.gui.apply_virtual_library(vl, update_tabs=False)

    def tab_moved(self, from_, to):
        self.current_db.new_api.set_pref('virt_libs_order', [str(self.tabData(i) or '') for i in range(self.count())])

    def tab_close(self, index):
        vl = str(self.tabData(index) or '')
        if vl:  # Don't allow closing the All Books tab
            self.current_db.new_api.set_pref('virt_libs_hidden', list(self.current_db.new_api.pref('virt_libs_hidden', ())) + [vl])
            self.removeTab(index)

    @property
    def current_db(self):
        return self.gui.current_db

    def rebuild(self):
        self.ignore_tab_changed = True
        try:
            self._rebuild()
        finally:
            self.ignore_tab_changed = False

    def _rebuild(self):
        db = self.current_db
        vl_map = db.new_api.pref('virtual_libraries', {})
        virt_libs = frozenset(vl_map)
        hidden = set(db.new_api.pref('virt_libs_hidden', ()))
        if hidden - virt_libs:
            hidden = hidden.intersection(virt_libs)
            db.new_api.set_pref('virt_libs_hidden', list(hidden))
        order = db.new_api.pref('virt_libs_order', ())
        while self.count():
            self.removeTab(0)
        current_lib = db.data.get_base_restriction_name()
        if current_lib in hidden:
            hidden.discard(current_lib)
            db.new_api.set_pref('virt_libs_hidden', list(hidden))
        current_idx = all_idx = None
        virt_libs = (set(virt_libs) - hidden) | {''}
        order = {x: i for i, x in enumerate(order)}
        for i, vl in enumerate(sorted(virt_libs, key=lambda x: (order.get(x, 0), sort_key(x)))):
            self.addTab(vl.replace('&', '&&') or _('All books'))
            sexp = vl_map.get(vl, None)
            if sexp is not None:
                self.setTabToolTip(i, _('Search expression for this Virtual library:') + '\n\n' + sexp)
            self.setTabData(i, vl)
            if vl == current_lib:
                current_idx = i
            if not vl:
                all_idx = i
        assert all_idx is not None
        self.setCurrentIndex(all_idx if current_idx is None else current_idx)
        if current_idx is None and current_lib:
            self.setTabText(all_idx, current_lib)
        btn = self.tabButton(all_idx, QTabBar.ButtonPosition.RightSide)
        if btn is not None:
            btn.setVisible(False)
        else:
            btn = self.tabButton(all_idx, QTabBar.ButtonPosition.LeftSide)
            if btn is not None:
                btn.setVisible(False)
        self.update_visibility()

    def update_current(self):
        self.rebuild()

    def contextMenuEvent(self, a0):
        m = QMenu(self)
        m.addAction(QIcon.ic('sort.png'), _('Sort tabs alphabetically'), self.sort_alphabetically)
        hidden = self.current_db.new_api.pref('virt_libs_hidden')
        if hidden:
            s = m.addMenu(_('Restore hidden tabs'))
            assert s is not None
            for x in hidden:
                s.addAction(x, partial(self.restore, x))
        m.addAction(_('Hide Virtual library tabs'), self.disable_bar)
        if gprefs['vl_tabs_closable']:
            m.addAction(QIcon.ic('drm-locked.png'), _('Lock Virtual library tabs'), self.lock_tab)
        else:
            m.addAction(QIcon.ic('drm-unlocked.png'), _('Unlock Virtual library tabs'), self.unlock_tab)
        i = self.tabAt(a0.pos())
        if i > -1:
            vl = str(self.tabData(i) or '')
            if vl:
                vln = vl.replace('&', '&&')
                m.addSeparator()
                m.addAction(QIcon.ic('edit_input.png'), _('Edit "%s"') % vln, partial(self.gui.do_create_edit, name=vl))
                m.addAction(QIcon.ic('trash.png'), _('Delete "%s"') % vln, partial(self.gui.remove_vl_triggered, name=vl))
        m.exec(a0.globalPos())

    def sort_alphabetically(self):
        self.current_db.new_api.set_pref('virt_libs_order', ())
        self.rebuild()

    def restore(self, x):
        h = self.current_db.new_api.pref('virt_libs_hidden', ())
        self.current_db.new_api.set_pref('virt_libs_hidden', list(set(h) - {x}))
        self.rebuild()


# }}}


class StatusBarButton(QToolButton):
    def __init__(self, parent, action_name, pref_name, on_click):
        super().__init__(parent=parent)
        act = parent.iactions[action_name]
        self.action_name = action_name
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.setAutoRaise(True)
        self.setIcon(QIcon.ic(act.action_spec[1]))
        self.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.setText(act.action_spec[0])
        self.setToolTip(act.action_spec[2])
        self.setVisible(gprefs[pref_name])
        parent.status_bar.addPermanentWidget(self)
        if on_click == 'menu':
            self.setMenu(act.qaction.menu())
        elif on_click == 'trigger':
            self.clicked.connect(act.qaction.trigger)
        else:
            raise ValueError(f'make_status_line_action_button: invalid on_click ({on_click}')


class LayoutMixin:  # {{{
    def place_layout_buttons(self: Main):
        if getattr(self, 'layout_buttons', None):
            for x in self.layout_buttons:
                self.status_bar.removeWidget(x)
        if self.layout_container.is_wide:
            self.button_order = 'sb', 'tb', 'cb', 'bs', 'gv', 'qv', 'bd'
        else:
            self.button_order = 'sb', 'tb', 'bd', 'gv', 'cb', 'bs', 'qv'
        self.layout_buttons = []
        s = self.style()
        assert s is not None
        stylename = str(s.objectName())
        for x in self.button_order:
            if x == 'gv':
                button = self.grid_view_button
            elif x == 'bs':
                button = self.bookshelf_view_button
            elif x == 'sb':
                button = self.search_bar_button
            else:
                button = self.layout_container.button_for(
                    {
                        'tb': 'tag_browser',
                        'bd': 'book_details',
                        'cb': 'cover_browser',
                        'qv': 'quick_view',
                    }[x]
                )
            self.layout_buttons.append(button)
            button.setVisible(gprefs['show_layout_buttons'])
            if ismacos and stylename != 'Calibre':
                button.setStyleSheet('''
                        QToolButton { background: none; border:none; padding: 0px; }
                        QToolButton:checked { background: rgba(0, 0, 0, 25%); }
                ''')
        for button in reversed(self.layout_buttons):
            self.status_bar.insertPermanentWidget(2, button)
        self.layout_button.setVisible(not gprefs['show_layout_buttons'])

    def init_layout_mixin(self: Main):
        self.vl_tabs = VLTabs(self)
        cl = self.centralwidget.layout()
        assert cl is not None
        cl.addWidget(self.vl_tabs)
        self.layout_container = CentralContainer(self)
        cl.addWidget(self.layout_container)
        self.book_details = BookDetails(self.layout_container.is_wide, self)
        self.stack = QStackedWidget(self)
        self.library_view = BooksView(self)
        self.library_view.setObjectName('library_view')
        stack = QStackedWidget(self)
        self.stack.addWidget(stack)
        av = self.library_view.alternate_views
        self.pin_container = av.set_stack(stack)
        self.grid_view = GridView(self)
        self.grid_view.setObjectName('grid_view')
        av.add_view('grid', self.grid_view)
        self.bookshelf_view = BookshelfView(self)
        self.bookshelf_view.setObjectName('bookshelf_view')
        av.add_view('bookshelf', self.bookshelf_view)
        self.tb_widget = TagBrowserWidget(self)
        self.memory_view = DeviceBooksView(self)
        self.stack.addWidget(self.memory_view)
        self.memory_view.setObjectName('memory_view')
        self.card_a_view = DeviceBooksView(self)
        self.stack.addWidget(self.card_a_view)
        self.card_a_view.setObjectName('card_a_view')
        self.card_b_view = DeviceBooksView(self)
        self.stack.addWidget(self.card_b_view)
        self.card_b_view.setObjectName('card_b_view')

        # This must use the base method to find the plugin because it hasn't
        # been fully initialized yet
        self.qv = find_plugin('Quickview')
        if self.qv and self.qv.actual_plugin_:
            self.qv = self.qv.actual_plugin_

        self.layout_container.initialize_with_gui(self, self.stack)
        self.layout_container.tag_browser_button.toggled.connect(self.tb_widget.set_pane_is_visible, Qt.ConnectionType.QueuedConnection)
        self.status_bar = StatusBar(self)
        self.grid_view_button = GridViewButton(self)
        self.bookshelf_view_button = BookshelfViewButton(self)
        self.search_bar_button = SearchBarButton(self)
        self.search_bar_button.toggled.connect(self.toggle_search_bar)

        self.layout_button = b = QToolButton(self)
        self.layout_button_menu = m = LayoutMenu(self)
        b.setAutoRaise(True), b.setCursor(Qt.CursorShape.PointingHandCursor)
        b.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        b.setText(_('Layout')), b.setIcon(QIcon.ic('layout.png'))
        b.setToolTip(_('Show and hide various parts of the calibre main window'))
        b.clicked.connect(m.toggle_visibility)
        self.status_bar.addPermanentWidget(b)

        # These must be after the layout button because it can be expanded into
        # the component buttons. Order: last is right-most.
        # The preferences status bar button isn't (yet) allowed on the status bar
        # self.sb_preferences_button = StatusBarButton(self, 'Preferences', 'show_sb_preference_button', 'trigger')
        self.sb_all_gui_actions_button = StatusBarButton(self, 'All GUI actions', 'show_sb_all_actions_button', 'menu')
        self.status_bar_extra_buttons = (self.sb_all_gui_actions_button,)

        self.status_bar.addPermanentWidget(self.jobs_button)
        self.setStatusBar(self.status_bar)
        self.status_bar.update_label.linkActivated.connect(self.update_link_clicked)

    def finalize_layout(self: Main):
        self.status_bar.initialize(self.system_tray_icon)
        self.book_details.show_book_info.connect(self.iactions['Show Book Details'].show_book_info)
        self.book_details.files_dropped.connect(self.iactions['Add Books'].files_dropped_on_book)
        self.book_details.cover_changed.connect(self.bd_cover_changed, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.open_cover_with.connect(self.bd_open_cover_with, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.open_fmt_with.connect(self.bd_open_fmt_with, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.edit_book.connect(self.bd_edit_book, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.cover_removed.connect(self.bd_cover_removed, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.remote_file_dropped.connect(self.iactions['Add Books'].remote_file_dropped_on_book, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.open_containing_folder.connect(self.iactions['View'].view_folder_for_id)
        self.book_details.open_data_folder.connect(self.iactions['View'].view_data_folder_for_id)
        self.book_details.view_specific_format.connect(self.iactions['View'].view_format_by_id)
        self.book_details.search_requested.connect(self.set_search_string_with_append)
        self.book_details.remove_specific_format.connect(self.iactions['Remove Books'].remove_format_by_id)
        self.book_details.remove_metadata_item.connect(self.iactions['Edit Metadata'].remove_metadata_item)
        self.book_details.save_specific_format.connect(self.iactions['Save To Disk'].save_library_format_by_ids)
        self.book_details.restore_specific_format.connect(self.iactions['Remove Books'].restore_format)
        self.book_details.set_cover_from_format.connect(self.iactions['Edit Metadata'].set_cover_from_format)
        self.book_details.copy_link.connect(self.bd_copy_link, type=Qt.ConnectionType.QueuedConnection)
        self.book_details.view_device_book.connect(self.iactions['View'].view_device_book)
        self.book_details.manage_category.connect(self.manage_category_triggerred)
        self.book_details.find_in_tag_browser.connect(self.find_in_tag_browser_triggered)
        self.book_details.edit_identifiers.connect(self.edit_identifiers_triggerred)
        self.book_details.compare_specific_format.connect(self.compare_format)

        m = self.library_view._model
        assert m is not None
        if m.rowCount(QModelIndex()) > 0:
            QTimer.singleShot(0, self.library_view.set_current_row)
            m.current_changed(self.library_view.currentIndex(), self.library_view.currentIndex())
        self.library_view.setFocus(Qt.FocusReason.OtherFocusReason)

    def show_panel(self, name):
        if name == 'search_bar':
            self.search_bar_button.setChecked(True)
            return
        self.layout_container.show_panel(name)

    def hide_panel(self, name):
        if name == 'search_bar':
            self.search_bar_button.setChecked(False)
            return
        self.layout_container.hide_panel(name)

    def set_search_string_with_append(self: Main, expression, append=''):
        current = self.search.text().strip()
        if append:
            expr = f'{current} {append} {expression}' if current else expression
        else:
            expr = expression
        self.search.set_search_string(expr)

    def edit_identifiers_triggerred(self: Main):
        book_id = self.library_view.current_book
        db = self.current_db.new_api
        identifiers = db.field_for('identifiers', book_id, default_value={})
        from calibre.gui2.metadata.basic_widgets import Identifiers

        d = Identifiers(identifiers, self)
        if d.exec() == QDialog.DialogCode.Accepted:
            identifiers = d.get_identifiers()
            db.set_field('identifiers', {book_id: identifiers})
            self.iactions['Edit Metadata'].refresh_books_after_metadata_edit({book_id})

    def manage_category_triggerred(self: Main, field, value):
        if field and value:
            if field == 'authors':
                self.do_author_sort_edit(self, value, select_sort=False, select_link=False, lookup_author=True)
            elif field:
                self.do_tags_list_edit(value, field)

    def find_in_tag_browser_triggered(self: Main, field, value):
        if field and value:
            tb = self.tb_widget
            tb.set_focus_to_find_box()
            le = tb.item_search.lineEdit()
            assert le is not None
            le.setText(field + ':=' + value)
            tb.do_find()

    def toggle_search_bar(self: Main, show):
        self.search_bar.setVisible(show)
        if show:
            self.search.setFocus(Qt.FocusReason.OtherFocusReason)

    def bd_cover_changed(self: Main, id_, cdata):
        self.current_db.set_cover(id_, cdata)
        self.refresh_cover_browser()

    def bd_open_cover_with(self: Main, book_id, entry):
        cpath = self.current_db.new_api.format_abspath(book_id, '__COVER_INTERNAL__')
        if cpath:
            if entry is None:
                pm = QPixmap()
                pm.load(cpath)
                pm.setDevicePixelRatio(self.devicePixelRatioF())
                if pm.isNull():
                    open_local_file(cpath)
                else:
                    from calibre.gui2.image_popup import ImageView

                    iv = ImageView(
                        qapplication_or_fail().focusWindow(),
                        pm,
                        QUrl.fromLocalFile(cpath),
                        geom_name='book_details_image_view',
                    )
                    iv(use_exec=True)
                return
            from calibre.gui2.open_with import run_program

            run_program(entry, cpath, self)

    def bd_open_fmt_with(self: Main, book_id, fmt, entry):
        path = self.current_db.new_api.format_abspath(book_id, fmt)
        if path:
            from calibre.gui2.open_with import run_program

            run_program(entry, path, self)
        else:
            fmt = fmt.upper()
            error_dialog(
                self,
                _('No %s format') % fmt,
                _('The book {0} does not have the {1} format').format(self.current_db.new_api.field_for('title', book_id, default_value=_('Unknown')), fmt),
                show=True,
            )

    def bd_edit_book(self: Main, book_id, fmt):
        from calibre.gui2.widgets import BusyCursor

        with BusyCursor():
            self.iactions['Tweak ePub'].ebook_edit_format(book_id, fmt)

    def open_with_action_triggerred(self: Main, fmt, entry, *args):
        book_id = self.library_view.current_book
        if book_id is not None:
            if fmt == 'cover_image':
                self.bd_open_cover_with(book_id, entry)
            else:
                self.bd_open_fmt_with(book_id, fmt, entry)

    def bd_cover_removed(self: Main, id_):
        self.current_db.remove_cover(id_, commit=True, notify=False)
        self.refresh_cover_browser()

    def bd_copy_link(self, url):
        if url:
            clipboard = QApplication.clipboard()
            assert clipboard is not None
            clipboard.setText(url)

    def compare_format(self: Main, book_id, fmt):
        db = self.current_db.new_api
        ofmt = fmt
        if fmt.startswith('ORIGINAL_'):
            fmt = fmt.partition('_')[-1]
        else:
            ofmt = 'ORIGINAL_' + fmt
        path1, path2 = db.format_abspath(book_id, ofmt), db.format_abspath(book_id, fmt)
        from calibre.gui2.tweak_book.diff.main import compare_books

        compare_books(
            path1,
            path2,
            parent=self,
            revert_msg=_('Restore %s') % ofmt,
            revert_callback=partial(self.iactions['Remove Books'].restore_format, book_id, ofmt),
            names=(ofmt, fmt),
        )

    def save_layout_state(self: Main):
        self.library_view.save_state()
        self.memory_view.save_state()
        self.card_a_view.save_state()
        self.card_b_view.save_state()
        self.layout_container.write_settings()
        self.grid_view_button.save_state()
        self.bookshelf_view_button.save_state()
        self.search_bar_button.save_state()

    def read_layout_settings(self: Main):
        # View states are restored automatically when set_database is called
        self.layout_container.read_settings()
        self.book_details.change_layout(self.layout_container.is_wide)
        self.place_layout_buttons()
        self.grid_view_button.restore_state()
        self.bookshelf_view_button.restore_state()
        self.search_bar_button.restore_state()

    def update_status_bar(self: Main, *args):
        v = self.current_view()
        selected = len(v.selectionModel().selectedRows())
        try:
            library_total, total, current = v.model().counts()
        except AttributeError:  # happens during shutdown
            if self.shutting_down:
                return
            raise
        self.status_bar.update_state(library_total, total, current, selected)


# }}}



# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

""""""

import struct

from calibre.ebooks.lrf import LRFParseError


class Tag:
    tags = {
        0x00: (6, '*ObjectStart'),
        0x01: (0, '*ObjectEnd'),
        0x02: (4, '*ObjectInfoLink'),
        0x03: (4, '*Link'),
        0x04: (4, '*StreamSize'),
        0x05: (0, '*StreamStart'),
        0x06: (0, '*StreamEnd'),
        0x07: (4, None),
        0x08: (4, None),
        0x09: (4, None),
        0x0A: (4, None),
        0x0B: ('type_one', '*ContainedObjectsList'),
        0x0D: (2, None),
        0x0E: (2, None),
        0x11: (2, None),
        0x12: (2, None),
        0x13: (2, None),
        0x14: (2, None),
        0x15: (2, None),
        0x16: ('string', None),
        0x17: (4, None),
        0x18: (4, None),
        0x19: (2, None),
        0x1A: (2, None),
        0x1B: (2, None),
        0x1C: (2, None),
        0x1D: (2, None),
        0x1E: (2, None),
        0x21: (2, None),
        0x22: (2, None),
        0x23: (2, None),
        0x24: (2, None),
        0x25: (2, None),
        0x26: (2, None),
        0x27: (2, None),
        0x28: (2, None),
        0x29: (6, None),
        0x2A: (2, None),
        0x2B: (2, None),
        0x2C: (2, None),
        0x2D: (4, None),
        0x2E: (2, None),
        0x31: (2, None),
        0x32: (2, None),
        0x33: (2, None),
        0x34: (4, None),
        0x35: (2, None),
        0x36: (2, None),
        0x37: (4, None),
        0x38: (2, None),
        0x39: (2, None),
        0x3A: (2, None),
        0x3C: (2, None),
        0x3D: (2, None),
        0x3E: (2, None),
        0x41: (2, None),
        0x42: (2, None),
        0x44: (4, None),
        0x45: (4, None),
        0x46: (2, None),
        0x47: (2, None),
        0x48: (2, None),
        0x49: (8, None),
        0x4A: (8, None),
        0x4B: (4, None),
        0x4C: (4, None),
        0x4D: (0, None),
        0x4E: (12, None),
        0x51: (2, None),
        0x52: (2, None),
        0x53: (4, None),
        0x54: (2, '*StreamFlags'),
        0x55: ('string', None),
        0x56: (2, None),
        0x57: (2, None),
        0x58: (2, None),
        0x59: ('string', None),
        0x5A: ('string', None),
        0x5B: (4, None),
        0x5C: ('type_one', None),
        0x5D: ('string', None),
        0x5E: (2, None),
        0x61: (2, None),
        0x62: (0, None),
        0x63: (0, None),
        0x64: (0, None),
        0x65: (0, None),
        0x66: (0, None),
        0x67: (0, None),
        0x68: (0, None),
        0x69: (0, None),
        0x6A: (0, None),
        0x6B: (0, None),
        0x6C: (8, None),
        0x6D: (2, None),
        0x6E: (0, None),
        0x71: (0, None),
        0x72: (0, None),
        0x73: (10, None),
        0x75: (2, None),
        0x76: (2, None),
        0x77: (2, None),
        0x78: ('tag_78', None),
        0x79: (2, None),
        0x7A: (2, None),
        0x7B: (4, None),
        0x7C: (4, '*ParentPageTree'),
        0x81: (0, None),
        0x82: (0, None),
        0xA1: (4, None),
        0xA2: (0, None),
        0xA5: ('unknown', None),
        0xA6: (0, None),
        0xA7: (4, None),
        0xA8: (0, None),
        0xA9: (0, None),
        0xAA: (0, None),
        0xAB: (0, None),
        0xAC: (0, None),
        0xAD: (0, None),
        0xAE: (0, None),
        0xB1: (0, None),
        0xB2: (0, None),
        0xB3: (0, None),
        0xB4: (0, None),
        0xB5: (0, None),
        0xB6: (0, None),
        0xB7: (0, None),
        0xB8: (0, None),
        0xB9: (0, None),
        0xBA: (0, None),
        0xBB: (0, None),
        0xBC: (0, None),
        0xBD: (0, None),
        0xBE: (0, None),
        0xC1: (0, None),
        0xC2: (0, None),
        0xC3: (2, None),
        0xC4: (0, None),
        0xC5: (2, None),
        0xC6: (2, None),
        0xC7: (0, None),
        0xC8: (2, None),
        0xC9: (0, None),
        0xCA: (2, None),
        0xCB: ('unknown', None),
        0xCC: (2, None),
        0xD1: (12, None),
        0xD2: (0, None),
        0xD4: (2, None),
        0xD6: (0, None),
        0xD7: (14, None),
        0xD8: (4, None),
        0xD9: (8, None),
        0xDA: (2, None),
        0xDB: (2, None),
        0xDC: (2, None),
        0xDD: (2, None),
        0xF1: (2, None),
        0xF2: (4, None),
        0xF3: (4, None),
        0xF4: (2, None),
        0xF5: (4, None),
        0xF6: (4, None),
        0xF7: (4, None),
        0xF8: (4, None),
        0xF9: (6, None),
    }
    name_map = {}
    for key in tags.keys():
        temp = tags[key][1]
        if temp is not None:
            name_map[key] = temp

    def __init__(self, stream):
        self.offset = stream.tell()
        tag_id = struct.unpack('<BB', stream.read(2))
        if tag_id[1] != 0xF5:
            raise LRFParseError(f'Bad tag ID {tag_id[1]:02X} at {self.offset}')
        if tag_id[0] not in self.__class__.tags:
            raise LRFParseError(f'Unknown tag ID: F5{tag_id[0]:02X}')

        self.id = 0xF500 + tag_id[0]

        size, self.name = self.__class__.tags[tag_id[0]]
        if isinstance(size, str):
            parser = getattr(self, size + '_parser')
            self.contents = parser(stream)
        else:
            self.contents = stream.read(size)

    def __str__(self):
        s = f'Tag {self.id:04X} '
        if self.name:
            s += self.name
        s += f' at {self.offset:08X}, contents: {self.contents!r}'
        return s

    @property
    def byte(self):
        if len(self.contents) != 1:
            raise LRFParseError(f'Bad parameter for tag ID: {self.id:04X}')
        return struct.unpack('<B', self.contents)[0]

    @property
    def word(self):
        if len(self.contents) != 2:
            raise LRFParseError(f'Bad parameter for tag ID: {self.id:04X}')
        return struct.unpack('<H', self.contents)[0]

    @property
    def sword(self):
        if len(self.contents) != 2:
            raise LRFParseError(f'Bad parameter for tag ID: {self.id:04X}')
        return struct.unpack('<h', self.contents)[0]

    @property
    def dword(self):
        if len(self.contents) != 4:
            raise LRFParseError(f'Bad parameter for tag ID: {self.id:04X}')
        return struct.unpack('<I', self.contents)[0]

    def dummy_parser(self, stream):
        raise LRFParseError(f'Unknown tag at {stream.tell():08X}')

    @classmethod
    def string_parser(self, stream):
        size = struct.unpack('<H', stream.read(2))[0]
        return str(stream.read(size), 'utf_16')

    def type_one_parser(self, stream):
        cnt = struct.unpack('<H', stream.read(2))[0]
        res = []
        while cnt > 0:
            res.append(struct.unpack('<I', stream.read(4))[0])
            cnt -= 1
        return res

    def tag_78_parser(self, stream):
        pos = stream.tell()
        res = []
        res.append(struct.unpack('<I', stream.read(4))[0])
        tag = Tag(stream)
        if tag.id != 0xF516:
            raise LRFParseError(f'Bad tag 78 at {pos:08X}')
        res.append(tag.contents)
        res.append(struct.unpack('<H', stream.read(2))[0])
        return res



#!/usr/bin/env python
# License: GPLv3 Copyright: 2014, Kovid Goyal <kovid at kovidgoyal.net>

import errno
import os
import shutil
import time
import traceback
from collections import defaultdict, namedtuple
from io import BytesIO
from queue import Empty
from typing import cast

from qt.core import QObject, Qt, pyqtSignal

from calibre import force_unicode, prints
from calibre.constants import DEBUG
from calibre.customize.ui import can_set_metadata
from calibre.db.constants import DATA_FILE_PATTERN
from calibre.db.errors import NoSuchFormat
from calibre.ebooks.metadata import authors_to_string
from calibre.ebooks.metadata.opf2 import metadata_to_opf
from calibre.gui2 import error_dialog, gprefs, open_local_file, warning_dialog
from calibre.gui2.dialogs.progress import ProgressDialog
from calibre.library.save_to_disk import find_plugboard, get_path_components, plugboard_save_to_disk_value, sanitize_args
from calibre.ptempfile import PersistentTemporaryDirectory, SpooledTemporaryFile
from calibre.utils.filenames import make_long_path_useable
from calibre.utils.ipc.pool import Failure, Pool
from calibre.utils.localization import _

BookId = namedtuple('BookId', 'title authors')


def ensure_unique_components(data):  # {{{
    cmap = defaultdict(set)
    bid_map = {}
    for book_id, (mi, components, fmts) in data.items():
        cmap[tuple(components)].add(book_id)
        bid_map[book_id] = components

    for book_ids in cmap.values():
        if len(book_ids) > 1:
            for i, book_id in enumerate(sorted(book_ids)[1:]):
                suffix = f' ({i + 1})'
                components = bid_map[book_id]
                components[-1] = components[-1] + suffix


# }}}


class SpooledFile(SpooledTemporaryFile):  # {{{
    def __init__(self, file_obj, max_size=50 * 1024 * 1024):
        self._file_obj = file_obj
        SpooledTemporaryFile.__init__(self, max_size)

    def rollover(self):
        if self._rolled:
            return
        orig = self._file
        newfile = self._file = self._file_obj
        del self._TemporaryFileArgs  # type: ignore

        newfile.write(cast(BytesIO, orig).getvalue())
        newfile.seek(orig.tell(), 0)

        self._rolled = True


# }}}


class Saver(QObject):
    do_one_signal = pyqtSignal()
    pd_title = ''

    def __init__(self, book_ids, db, opts, root, parent=None, pool=None):
        QObject.__init__(self, parent)
        self.db = db.new_api
        self.plugboards = self.db.pref('plugboards', {})
        self.template_functions = self.db.pref('user_template_functions', [])
        self.library_id = self.db.library_id
        # This call to load_user_template_functions isn't needed because
        # __init__ is running on the GUI thread. It must be done in the separate
        # process by the worker
        # from calibre.gui2 import is_gui_thread
        # print(f'Saver __init__ is_gui_thread: {is_gui_thread()}')
        # load_user_template_functions('', self.template_functions)
        self.collected_data = {}
        self.errors = defaultdict(list)
        self._book_id_data = {}
        self.all_book_ids = frozenset(book_ids)
        self.pd = ProgressDialog(
            self.pd_title or (_('Saving %d books...') % len(self.all_book_ids)),
            _('Collecting metadata...'),
            min=0,
            max=0,
            parent=parent,
            icon='save.png',
        )
        self.do_one_signal.connect(self.tick, type=Qt.ConnectionType.QueuedConnection)
        self.do_one = self.do_one_collect
        self.ids_to_collect = iter(self.all_book_ids)
        self.tdir = PersistentTemporaryDirectory('_save_to_disk')
        self.pool = pool

        self.pd.show()
        self.root, self.opts, self.path_length = sanitize_args(root, opts)
        self.do_one_signal.emit()
        if DEBUG:
            self.start_time = time.time()

    def tick(self):
        pd = self.pd
        assert pd is not None
        if pd.canceled:
            pd.close()
            pd.deleteLater()
            self.break_cycles()
            return
        self.do_one()

    def break_cycles(self):
        shutil.rmtree(self.tdir, ignore_errors=True)
        if self.pool is not None:
            self.pool.shutdown()
        self.setParent(None)
        self.jobs = self.pool = self.plugboards = self.template_functions = self.library_id = self.collected_data = self.all_book_ids = self.pd = self.db = None
        self.deleteLater()

    def book_id_data(self, book_id):
        ans = self._book_id_data.get(book_id)
        if ans is None:
            try:
                db = self.db
                assert db is not None
                ans = BookId(db.field_for('title', book_id), db.field_for('authors', book_id))
            except Exception:
                ans = BookId((_('Unknown') + f' ({book_id})'), (_('Unknown'),))
            self._book_id_data[book_id] = ans
        return ans

    def do_one_collect(self):
        try:
            book_id = next(self.ids_to_collect)
        except StopIteration:
            self.collection_finished()
            return
        try:
            self.collect_data(book_id)
        except Exception:
            self.errors[book_id].append(('critical', traceback.format_exc()))
        self.do_one_signal.emit()

    def collect_data(self, book_id):
        db = self.db
        assert db is not None
        mi = db.get_metadata(book_id)
        self._book_id_data[book_id] = BookId(mi.title, mi.authors)
        components = get_path_components(self.opts, mi, book_id, self.path_length)
        collected_data = self.collected_data
        assert collected_data is not None
        collected_data[book_id] = (mi, components, {fmt.lower() for fmt in db.formats(book_id)})

    def collection_finished(self):
        self.do_one = self.do_one_write
        collected_data = self.collected_data
        assert collected_data is not None
        ensure_unique_components(collected_data)
        self.ids_to_write = iter(collected_data)
        pd = self.pd
        assert pd is not None
        pd.title = _('Copying files and writing metadata...') if self.opts.update_metadata else _('Copying files...')
        pd.max = len(collected_data)
        pd.value = 0
        if self.opts.update_metadata:
            all_fmts = {fmt for data in collected_data.values() for fmt in data[2]}
            plugboards_cache = {fmt: find_plugboard(plugboard_save_to_disk_value, fmt, self.plugboards) for fmt in all_fmts}
            self.pool = Pool(name='SaveToDisk') if self.pool is None else self.pool
            try:
                self.pool.set_common_data({
                    'plugboard_cache': plugboards_cache,
                    'template_functions': self.template_functions,
                    'library_id': self.library_id,
                })
            except Failure as err:
                error_dialog(
                    pd,
                    _('Critical failure'),
                    _('Could not save books to disk, click "Show details" for more information'),
                    det_msg=force_unicode(err.failure_message) + '\n' + force_unicode(err.details),
                    show=True,
                )
                pd.canceled = True
        self.do_one_signal.emit()

    def do_one_write(self):
        try:
            book_id = next(self.ids_to_write)
        except StopIteration:
            self.writing_finished()
            return
        if not self.opts.update_metadata:
            pd = self.pd
            assert pd is not None
            pd.msg = self.book_id_data(book_id).title
            pd.value += 1
        collected_data = self.collected_data
        assert collected_data is not None
        try:
            self.write_book(book_id, *collected_data[book_id])
        except Exception:
            self.errors[book_id].append(('critical', traceback.format_exc()))
        self.consume_results()
        self.do_one_signal.emit()

    def consume_results(self):
        if self.pool is not None:
            while True:
                try:
                    worker_result = self.pool.results.get_nowait()
                except Empty:
                    break
                book_id = worker_result.id
                pd = self.pd
                assert pd is not None
                if worker_result.is_terminal_failure:
                    error_dialog(
                        pd,
                        _('Critical failure'),
                        _('The update metadata worker process crashed while processing the book %s. Saving is aborted.') % self.book_id_data(book_id).title,
                        show=True,
                    )
                    pd.canceled = True
                    return
                result = worker_result.result
                pd.value += 1
                pd.msg = self.book_id_data(book_id).title
                if result.err is not None:
                    self.errors[book_id].append(('metadata', (None, result.err + '\n' + result.traceback)))
                if result.value:
                    for fmt, tb in result.value:
                        self.errors[book_id].append(('metadata', (fmt, tb)))

    def write_book(self, book_id, mi, components, fmts):
        db = self.db
        assert db is not None
        base_path = os.path.join(self.root, *components)
        base_dir = os.path.dirname(base_path)
        if self.opts.formats and self.opts.formats != 'all':
            if self.opts.formats == '..cover..':
                fmts = set()
            else:
                asked_formats = {x.lower().strip() for x in self.opts.formats.split(',')}
                fmts = asked_formats.intersection(fmts)
                if not fmts:
                    self.errors[book_id].append(('critical', _('Requested formats not available')))
                    return

        extra_files = {}
        if self.opts.save_extra_files:
            extra_files = {}
            for efx in db.new_api.list_extra_files(int(book_id), pattern=DATA_FILE_PATTERN):
                extra_files[efx.relpath] = efx.file_path
        if not fmts and not self.opts.write_opf and not self.opts.save_cover and not extra_files:
            return

        # On windows python incorrectly raises an access denied exception
        # when trying to create the root of a drive, like C:\
        if os.path.dirname(base_dir) != base_dir:
            try:
                os.makedirs(base_dir)
            except OSError as err:
                if err.errno != errno.EEXIST:
                    raise

        if self.opts.update_metadata:
            d = {}
            d['last_modified'] = mi.last_modified.isoformat()

        cdata = db.cover(book_id)
        mi.cover, mi.cover_data = None, (None, None)

        if cdata:
            fname = None
            if self.opts.save_cover:
                fname = base_path + os.extsep + 'jpg'
                mi.cover = os.path.basename(fname)
            elif self.opts.update_metadata:
                fname = os.path.join(self.tdir, f'{book_id}.jpg')

            if fname:
                with open(fname, 'wb') as f:
                    f.write(cdata)
                if self.opts.update_metadata:
                    d['cover'] = fname

        fname = None
        if self.opts.write_opf:
            fname = base_path + os.extsep + 'opf'
        elif self.opts.update_metadata:
            fname = os.path.join(self.tdir, f'{book_id}.opf')
        if fname:
            opf = metadata_to_opf(mi)
            with open(fname, 'wb') as f:
                f.write(opf)
            if self.opts.update_metadata:
                d['opf'] = fname
        mi.cover, mi.cover_data = None, (None, None)
        fmts_list: list[str] = []
        if self.opts.update_metadata:
            d['fmts'] = fmts_list
        if extra_files:
            for relpath, src_path in extra_files.items():
                src_path = make_long_path_useable(src_path)
                if os.access(src_path, os.R_OK):
                    dest = make_long_path_useable(os.path.abspath(os.path.join(base_dir, relpath)))
                    try:
                        shutil.copy2(src_path, dest)
                    except FileNotFoundError:
                        os.makedirs(os.path.dirname(dest), exist_ok=True)
                        shutil.copy2(src_path, dest)
        for fmt in fmts:
            try:
                fmtpath = self.write_fmt(book_id, fmt, base_path)
                if fmtpath and self.opts.update_metadata and can_set_metadata(fmt):
                    fmts_list.append(fmtpath)
            except Exception:
                self.errors[book_id].append(('fmt', (fmt, traceback.format_exc())))
        if self.opts.update_metadata:
            if fmts_list:
                pool = self.pool
                assert pool is not None
                pd = self.pd
                assert pd is not None
                try:
                    pool(book_id, 'calibre.library.save_to_disk', 'update_serialized_metadata', d)
                except Failure as err:
                    error_dialog(
                        pd,
                        _('Critical failure'),
                        _('Could not save books to disk, click "Show details" for more information'),
                        det_msg=str(err.failure_message) + '\n' + str(err.details),
                        show=True,
                    )
                    pd.canceled = True
            else:
                pd = self.pd
                assert pd is not None
                pd.value += 1
                pd.msg = self.book_id_data(book_id).title

    def write_fmt(self, book_id, fmt, base_path):
        db = self.db
        assert db is not None
        fmtpath = base_path + os.extsep + fmt
        written = False
        with open(fmtpath, 'w+b') as f:
            try:
                db.copy_format_to(book_id, fmt, f)
                written = True
            except NoSuchFormat:
                self.errors[book_id].append(('fmt', (fmt, _('No %s format file present') % fmt.upper())))
        if not written:
            os.remove(fmtpath)
        if written:
            return fmtpath

    def writing_finished(self):
        if not self.opts.update_metadata:
            self.updating_metadata_finished()
        else:
            self.do_one = self.do_one_update
            self.do_one_signal.emit()

    def do_one_update(self):
        self.consume_results()
        pool = self.pool
        assert pool is not None
        pd = self.pd
        assert pd is not None
        try:
            pool.wait_for_tasks(0.1)
        except Failure as err:
            error_dialog(
                pd,
                _('Critical failure'),
                _('Could not save books to disk, click "Show details" for more information'),
                det_msg=str(err.failure_message) + '\n' + str(err.details),
                show=True,
            )
            pd.canceled = True
        except RuntimeError:
            pass  # tasks not completed
        else:
            self.consume_results()
            return self.updating_metadata_finished()
        self.do_one_signal.emit()

    def updating_metadata_finished(self):
        if DEBUG:
            all_book_ids = self.all_book_ids
            assert all_book_ids is not None
            prints(f'Saved {len(all_book_ids)} books in {time.time() - self.start_time:.1f} seconds')
        pd = self.pd
        assert pd is not None
        pd.close()
        pd.deleteLater()
        self.report()
        self.break_cycles()
        self.on_complete()

    def on_complete(self):
        if gprefs['show_files_after_save']:
            open_local_file(self.root)

    def format_report(self):
        report = []
        a = report.append

        def indent(text):
            text = force_unicode(text)
            return '\xa0\xa0\xa0\xa0' + '\n\xa0\xa0\xa0\xa0'.join(text.splitlines())

        for book_id, errors in self.errors.items():
            types = {t for t, data in errors}
            title, authors = self.book_id_data(book_id).title, authors_to_string(self.book_id_data(book_id).authors[:1])
            if report:
                a('\n' + ('_' * 70) + '\n')
            if 'critical' in types:
                a(_('Failed to save: {0} by {1} to disk, with error:').format(title, authors))
                for t, tb in errors:
                    if t == 'critical':
                        a(indent(tb))
            else:
                errs = defaultdict(list)
                for t, data in errors:
                    errs[t].append(data)
                for fmt, tb in errs['fmt']:
                    a(_('Failed to save the {2} format of: {0} by {1} to disk, with error:').format(title, authors, fmt.upper()))
                    a(indent(tb)), a('')
                for fmt, tb in errs['metadata']:
                    if fmt:
                        a(_('Failed to update the metadata in the {2} format of: {0} by {1}, with error:').format(title, authors, fmt.upper()))
                    else:
                        a(_('Failed to update the metadata in all formats of: {0} by {1}, with error:').format(title, authors))
                    a(indent(tb)), a('')
        return '\n'.join(report)

    def report(self):
        if not self.errors:
            return
        all_book_ids = self.all_book_ids
        assert all_book_ids is not None
        err_types = {e[0] for errors in self.errors.values() for e in errors}
        if err_types == {'metadata'}:
            msg = _('Failed to update metadata in some books, click "Show details" for more information')
            d = warning_dialog
        elif len(self.errors) == len(all_book_ids):
            msg = _('Failed to save any books to disk, click "Show details" for more information')
            d = error_dialog
        else:
            msg = _('Failed to save some books to disk, click "Show details" for more information')
            d = warning_dialog
        d(self.parent(), _('Error while saving'), msg, det_msg=self.format_report(), show=True)



# License: GPLv3 Copyright: 2010, Hiroshi Miura <miurahr@linux.com>

"""
Decode unicode text to an ASCII representation of the text in Vietnamese.

"""

from calibre.ebooks.unihandecode.unicodepoints import CODEPOINTS
from calibre.ebooks.unihandecode.unidecoder import Unidecoder
from calibre.ebooks.unihandecode.vncodepoints import CODEPOINTS as HANCODES


class Vndecoder(Unidecoder):
    codepoints = {}

    def __init__(self):
        self.codepoints = CODEPOINTS
        self.codepoints.update(HANCODES)



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

from calibre.ebooks.conversion.config import OPTIONS
from calibre.gui2.convert import Widget
from calibre.gui2.convert.htmlz_output_ui import Ui_Form
from calibre.utils.localization import _

format_model = None


class PluginWidget(Widget, Ui_Form):
    TITLE = _('HTMLZ output')
    HELP = _('Options specific to') + ' HTMLZ ' + _('output')
    COMMIT_NAME = 'htmlz_output'
    ICON = 'mimetypes/html.png'

    def __init__(self, parent, get_option, get_help, db=None, book_id=None):
        Widget.__init__(self, parent, OPTIONS['output']['htmlz'])
        self.db, self.book_id = db, book_id
        for x in get_option('htmlz_css_type').option.choices:
            self.opt_htmlz_css_type.addItem(x)
        for x in get_option('htmlz_class_style').option.choices:
            self.opt_htmlz_class_style.addItem(x)
        self.initialize_options(get_option, get_help, db, book_id)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>

EXT_MAP = {
    'default': 'unknown',
    'dir': 'dir',
    'zero': 'zero',
    'jpeg': 'jpeg',
    'jpg': 'jpeg',
    'gif': 'gif',
    'png': 'png',
    'bmp': 'bmp',
    'cbz': 'cbz',
    'cbr': 'cbr',
    'svg': 'svg',
    'html': 'html',
    'htmlz': 'html',
    'htm': 'html',
    'xhtml': 'html',
    'xhtm': 'html',
    'lit': 'lit',
    'lrf': 'lrf',
    'lrx': 'lrx',
    'pdf': 'pdf',
    'pdr': 'zero',
    'rar': 'rar',
    'zip': 'zip',
    'txt': 'txt',
    'text': 'txt',
    'prc': 'mobi',
    'azw': 'mobi',
    'mobi': 'mobi',
    'pobi': 'mobi',
    'mbp': 'zero',
    'azw1': 'tpz',
    'azw2': 'azw2',
    'azw3': 'azw3',
    'kfx': 'kfx',
    'kfx-zip': 'kfx',
    'azw4': 'pdf',
    'tpz': 'tpz',
    'tan': 'zero',
    'epub': 'epub',
    'fb2': 'fb2',
    'rtf': 'rtf',
    'odt': 'odt',
    'snb': 'snb',
    'djv': 'djvu',
    'djvu': 'djvu',
    'xps': 'xps',
    'oxps': 'xps',
    'docx': 'docx',
    'opml': 'opml',
}



# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _


class USER_DEFINED(USBMS):
    name = 'User Defined USB driver'
    gui_name = 'User Defined USB Device'
    author = 'Kovid Goyal'
    supported_platforms = ['windows', 'osx', 'linux']

    # Ordered list of supported formats
    FORMATS = ['epub', 'mobi', 'pdf']

    VENDOR_ID = 0xFFFF
    PRODUCT_ID = 0xFFFF
    BCD = None

    EBOOK_DIR_MAIN = ''
    EBOOK_DIR_CARD_A = ''

    VENDOR_NAME = []
    WINDOWS_MAIN_MEM = ''
    WINDOWS_CARD_A_MEM = ''

    OSX_MAIN_MEM = 'Device Main Memory'

    MAIN_MEMORY_VOLUME_LABEL = 'Device Main Memory'

    SUPPORTS_SUB_DIRS = True

    EXTRA_CUSTOMIZATION_MESSAGE = [
        _('USB Vendor ID (in hex)') + ':::<p>' + _('Get this ID using Preferences -> Misc -> Get information to set up the user-defined device') + '</p>',
        _('USB Product ID (in hex)') + ':::<p>' + _('Get this ID using Preferences -> Misc -> Get information to set up the user-defined device') + '</p>',
        _('USB Revision ID (in hex)') + ':::<p>' + _('Get this ID using Preferences -> Misc -> Get information to set up the user-defined device') + '</p>',
        '',
        _('Unused (leave blank)') + ':::<p>' + _('This field is no longer used, leave it blank.') + '</p>',
        _('Unused (leave blank)') + ':::<p>' + _('This field is no longer used, leave it blank.') + '</p>',
        _('Unused (leave blank)') + ':::<p>' + _('This field is no longer used, leave it blank.') + '</p>',
        _('Unused (leave blank)') + ':::<p>' + _('This field is no longer used, leave it blank.') + '</p>',
        _('Main memory folder')
        + ':::<p>'
        + _('Enter the folder where the books are to be stored. This folder is prepended to any send_to_device template')
        + '</p>',
        _('Card A folder')
        + ':::<p>'
        + _('Enter the folder where the books are to be stored. This folder is prepended to any send_to_device template')
        + '</p>',
        _('Swap main and card A')
        + ':::<p>'
        + _("Check this box if the device's main memory is being seen as card a and the card is being seen as main memory")
        + '</p>',
    ]
    EXTRA_CUSTOMIZATION_DEFAULT = [
        '0xffff',
        '0xffff',
        '0xffff',
        None,
        '',
        '',
        '',
        '',
        '',
        '',
        False,
    ]
    OPT_USB_VENDOR_ID = 0
    OPT_USB_PRODUCT_ID = 1
    OPT_USB_REVISION_ID = 2
    # OPT 3 isn't used
    OPT_USB_WINDOWS_MM_VEN_ID = 4
    OPT_USB_WINDOWS_MM_ID = 5
    OPT_USB_WINDOWS_CA_VEN_ID = 6
    OPT_USB_WINDOWS_CA_ID = 7
    OPT_MAIN_MEM_FOLDER = 8
    OPT_CARD_A_FOLDER = 9
    OPT_SWAP_MAIN_AND_CARD = 10

    def initialize(self):
        self.plugin_needs_delayed_initialization = True
        USBMS.initialize(self)

    def do_delayed_plugin_initialization(self):
        try:
            e = self.settings().extra_customization
            if e[self.OPT_USB_VENDOR_ID]:
                self.VENDOR_ID = int(e[self.OPT_USB_VENDOR_ID], 16)
                self.PRODUCT_ID = int(e[self.OPT_USB_PRODUCT_ID], 16)
                self.BCD = [int(e[self.OPT_USB_REVISION_ID], 16)]
                if e[self.OPT_USB_WINDOWS_MM_VEN_ID]:
                    self.VENDOR_NAME.append(e[self.OPT_USB_WINDOWS_MM_VEN_ID])
                if e[self.OPT_USB_WINDOWS_CA_VEN_ID] and e[self.OPT_USB_WINDOWS_CA_VEN_ID] not in self.VENDOR_NAME:
                    self.VENDOR_NAME.append(e[self.OPT_USB_WINDOWS_CA_VEN_ID])
                self.WINDOWS_MAIN_MEM = e[self.OPT_USB_WINDOWS_MM_ID] + '&'
                self.WINDOWS_CARD_A_MEM = e[self.OPT_USB_WINDOWS_CA_ID] + '&'
                self.EBOOK_DIR_MAIN = e[self.OPT_MAIN_MEM_FOLDER]
                self.EBOOK_DIR_CARD_A = e[self.OPT_CARD_A_FOLDER]
        except Exception:
            import traceback

            traceback.print_exc()
        self.plugin_needs_delayed_initialization = False

    def windows_sort_drives(self, drives):
        if len(drives) < 2:
            return drives
        e = self.settings().extra_customization
        if not e[self.OPT_SWAP_MAIN_AND_CARD]:
            return drives
        main = drives.get('main', None)
        carda = drives.get('carda', None)
        if main and carda:
            drives['main'] = carda
            drives['carda'] = main
        return drives

    def linux_swap_drives(self, drives):
        if len(drives) < 2 or not drives[0] or not drives[1]:
            return drives
        e = self.settings().extra_customization
        if not e[self.OPT_SWAP_MAIN_AND_CARD]:
            return drives
        drives = list(drives)
        t = drives[0]
        drives[0] = drives[1]
        drives[1] = t
        return tuple(drives)

    def osx_sort_names(self, names):
        if len(names) < 2:
            return names
        e = self.settings().extra_customization
        if not e[self.OPT_SWAP_MAIN_AND_CARD]:
            return names
        main = names.get('main', None)
        card = names.get('carda', None)

        if main is not None and card is not None:
            names['main'] = card
            names['carda'] = main

        return names



# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

import io
import os

from calibre import fsync
from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _
from calibre.utils.resources import get_image_path as I

HTC_BCDS = [0x100, 0x0222, 0x0224, 0x0226, 0x227, 0x228, 0x229, 0x0231, 0x9999]


class ANDROID(USBMS):
    name = 'Android driver'
    gui_name = 'Android phone'
    description = _('Communicate with Android phones.')
    author = 'Kovid Goyal'
    supported_platforms = ['windows', 'osx', 'linux']

    # Ordered list of supported formats
    FORMATS = ['epub', 'pdf']

    VENDOR_ID = {
        # HTC
        0x0BB4: {
            0xC02: HTC_BCDS,
            0xC01: HTC_BCDS,
            0xFF9: HTC_BCDS,
            0xC86: HTC_BCDS,
            0xC87: HTC_BCDS,
            0xC8D: HTC_BCDS,
            0xC91: HTC_BCDS,
            0xC92: HTC_BCDS,
            0xC97: HTC_BCDS,
            0xC99: HTC_BCDS,
            0xCA2: HTC_BCDS,
            0xCA3: HTC_BCDS,
            0xCA4: HTC_BCDS,
            0xCA9: HTC_BCDS,
            0xCAC: HTC_BCDS,
            0xCBA: HTC_BCDS,
            0xCCF: HTC_BCDS,
            0xCD6: HTC_BCDS,
            0xCE5: HTC_BCDS,
            0xCEC: HTC_BCDS,
            0x0CF5: HTC_BCDS,
            0x2910: HTC_BCDS,
            0xE77: HTC_BCDS,
            0x0001: [0x255],
        },
        # Eken
        0x040D: {0x8510: [0x0001], 0x0851: [0x1]},
        # Trekstor
        0x1E68: {
            0x006A: [0x0231],
            0x0062: [0x222],  # Surftab ventos https://bugs.launchpad.net/bugs/1204885
        },
        # Motorola
        0x22B8: {
            0x41D9: [0x216],
            0x2D61: [0x100],
            0x2D67: [0x100],
            0x2DE8: [0x229],
            0x41DB: [0x216],
            0x4285: [0x216],
            0x42A3: [0x216],
            0x4286: [0x216],
            0x42B3: [0x216],
            0x42B4: [0x216],
            0x7086: [0x0226],
            0x70A8: [0x9999],
            0x42C4: [0x216],
            0x70C6: [0x226],
            0x70C7: [0x226],
            0x4316: [0x216],
            0x4317: [0x216],
            0x42D6: [0x216],
            0x42D7: [0x216],
            0x42F7: [0x216],
            0x4365: [0x216],
            0x4366: [0x216],
            0x4371: [0x216],
        },
        # Freescale
        0x15A2: {0x0C01: [0x226]},
        # Alcatel
        0x05C6: {
            0x9018: [0x0226],
        },
        # Sony Ericsson
        0xFCE: {
            0xA173: [0x216],
            0xD12E: [0x0100],
            0xE156: [0x226],
            0xE15D: [0x226],
            0xE14F: [0x0226],
            0x614F: [0x0226, 0x100],
            0x6156: [0x0226, 0x100],
        },
        # Google
        0x18D1: {
            0x0001: [0x0222, 0x0223, 0x230, 0x255, 0x9999],
            0x0002: [0x9999],
            0x0003: [0x0230, 0x9999],
            0x4E11: [0x0100, 0x226, 0x227],
            0x4E12: [0x0100, 0x226, 0x227],
            0x4E21: [0x0100, 0x226, 0x227, 0x231],
            0x4E22: [0x0100, 0x226, 0x227, 0x231],
            0xB058: [0x0222, 0x226, 0x227],
            0x0FF9: [0x0226],
            0xC91: HTC_BCDS,
            0xDDDD: [0x216],
            0x0D01: [0x9999],
            0x0D02: [0x9999],
            0x2D01: [0x9999],
            0xDEED: [0x231, 0x226],
        },
        # Samsung
        0x04E8: {
            0x681D: [0x0222, 0x0223, 0x0224, 0x0400],
            0x681C: [0x0222, 0x0223, 0x0224, 0x0400],
            0x6640: [0x0100],
            0x685B: [0x0400, 0x0226],
            0x685E: [0x0400, 0x226],
            0x6860: [0x0400],
            0x6863: [0x226],
            0x6877: [0x0400],
            0x689E: [0x0400],
            0xDEED: [0x0222],
            0x1234: [0x0400],
        },
        # Viewsonic/Vizio
        0x0489: {
            0xC000: [0x0226],
            0xC001: [0x0226],
            0xC004: [0x0226],
            0x8801: [0x0226, 0x0227],
            0xE115: [0x0216],  # PocketBook A10
        },
        # Another Viewsonic
        0x0BB0: {
            0x2A2B: [0x0226, 0x0227],
        },
        # Acer
        0x502: {0x3203: [0x0100, 0x224]},
        # Dell
        0x413C: {0xB007: [0x0100, 0x0224, 0x0226]},
        # LG
        0x1004: {
            0x61C5: [0x100, 0x226, 0x227, 0x229, 0x9999],
            0x61CC: [0x226, 0x227, 0x9999, 0x100],
            0x61CE: [0x226, 0x227, 0x9999, 0x100],
            0x618E: [0x226, 0x227, 0x9999, 0x100],
            0x6205: [0x226, 0x227, 0x9999, 0x100],
            0x6234: [0x231],
        },
        # Archos
        0x0E79: {
            0x1400: [0x0222, 0x0216],
            0x1408: [0x0222, 0x0216],
            0x1411: [0x216],
            0x1417: [0x0216],
            0x1419: [0x0216],
            0x1420: [0x0216],
            0x1422: [0x0216],
        },
        # Huawei
        # Disabled as this USB id is used by various USB flash drives
        # 0x45e : { 0x00e1 : [0x007], },
        # T-Mobile
        0x0408: {0x03BA: [0x0109]},
        # Xperia
        0x13D3: {0x3304: [0x0001, 0x0002]},
        # ZTE
        0x19D2: {0x1353: [0x226], 0x1351: [0x227]},
        # Advent
        0x0955: {0x7100: [0x9999]},  # This is the same as the Notion Ink Adam
        # Kobo
        0x2237: {0x2208: [0x0226]},
        # Lenovo
        0x17EF: {
            0x7421: [0x0216],
            0x741B: [0x9999],
            0x7640: [0x0255],
        },
        # Pantech
        0x10A9: {0x6050: [0x227]},
        # Prestigio and Teclast
        0x2207: {0: [0x222], 0x10: [0x222]},
        # OPPO
        0x22D9: {0x2768: [0x228]},
    }
    EBOOK_DIR_MAIN = ['eBooks/import', 'wordplayer/calibretransfer', 'Books', 'sdcard/ebooks']
    EXTRA_CUSTOMIZATION_MESSAGE = [
        _("Comma separated list of folders to send e-books to on the device's <b>main memory</b>. The first one that exists will be used"),
        _("Comma separated list of folders to send e-books to on the device's <b>storage cards</b>. The first one that exists will be used"),
    ]

    EXTRA_CUSTOMIZATION_DEFAULT = [', '.join(EBOOK_DIR_MAIN), '']

    VENDOR_NAME = [
        'HTC',
        'MOTOROLA',
        'GOOGLE_',
        'ANDROID',
        'ACER',
        'GT-I5700',
        'SAMSUNG',
        'DELL',
        'LINUX',
        'GOOGLE',
        'ARCHOS',
        'TELECHIP',
        'HUAWEI',
        'T-MOBILE',
        'SEMC',
        'LGE',
        'NVIDIA',
        'GENERIC-',
        'ZTE',
        'MID',
        'QUALCOMM',
        'PANDIGIT',
        'HYSTON',
        'VIZIO',
        'GOOGLE',
        'FREESCAL',
        'KOBO_INC',
        'LENOVO',
        'ROCKCHIP',
        'POCKET',
        'ONDA_MID',
        'ZENITHIN',
        'INGENIC',
        'PMID701C',
        'PD',
        'PMP5097C',
        'MASS',
        'NOVO7',
        'ZEKI',
        'COBY',
        'SXZ',
        'USB_2.0',
        'COBY_MID',
        'VS',
        'AINOL',
        'TOPWISE',
        'PAD703',
        'NEXT8D12',
        'MEDIATEK',
        'KEENHI',
        'TECLAST',
        'SURFTAB',
        'XENTA',
        'OBREEY_S',
        'SURFTAB_',
        'ONYX-INT',
        'IMCOSYS',
        'SURFPAD3',
        'GRAMMATA',
    ]
    WINDOWS_MAIN_MEM = [
        'ANDROID_PHONE',
        'A855',
        'A853',
        'A953',
        'INC.NEXUS_ONE',
        '__UMS_COMPOSITE',
        '_MB200',
        'MASS_STORAGE',
        '_-_CARD',
        'SGH-I897',
        'GT-I9000',
        'FILE-STOR_GADGET',
        'SGH-T959_CARD',
        'SGH-T959',
        'SAMSUNG_ANDROID',
        'SCH-I500_CARD',
        'SPH-D700_CARD',
        'MB810',
        'GT-P1000',
        'DESIRE',
        'SGH-T849',
        '_MB300',
        'A70S',
        'S_ANDROID',
        'A101IT',
        'A70H',
        'IDEOS_TABLET',
        'MYTOUCH_4G',
        'UMS_COMPOSITE',
        'SCH-I800_CARD',
        '7',
        'A956',
        'A955',
        'A43',
        'ANDROID_PLATFORM',
        'TEGRA_2',
        'MB860',
        'MULTI-CARD',
        'MID7015A',
        'INCREDIBLE',
        'A7EB',
        'STREAK',
        'MB525',
        'ANDROID2.3',
        'SGH-I997',
        'GT-I5800_CARD',
        'MB612',
        'GT-S5830_CARD',
        'GT-S5570_CARD',
        'MB870',
        'MID7015A',
        'ALPANDIGITAL',
        'ANDROID_MID',
        'VTAB1008',
        'EMX51_BBG_ANDROI',
        'UMS',
        '.K080',
        'P990',
        'LTE',
        'MB853',
        'GT-S5660_CARD',
        'A107',
        'GT-I9003_CARD',
        'XT912',
        'FILE-CD_GADGET',
        'RK29_SDK',
        'MB855',
        'XT910',
        'BOOK_A10',
        'USB_2.0_DRIVER',
        'I9100T',
        'P999DW',
        'KTABLET_PC',
        'INGENIC',
        'GT-I9001_CARD',
        'USB_2.0',
        'GT-S5830L_CARD',
        'UNIVERSE',
        'XT875',
        'PRO',
        '.KOBO_VOX',
        'THINKPAD_TABLET',
        'SGH-T989',
        'YP-G70',
        'STORAGE_DEVICE',
        'ADVANCED',
        'SGH-I727',
        'USB_FLASH_DRIVER',
        'ANDROID',
        'S5830I_CARD',
        'MID7042',
        'LINK-CREATE',
        '7035',
        'VIEWPAD_7E',
        'NOVO7',
        'MB526',
        '_USB#WYK7MSF8KE',
        'TABLET_PC',
        'F',
        'MT65XX_MS',
        'ICS',
        'E400',
        '__FILE-STOR_GADG',
        'ST80208-1',
        'GT-S5660M_CARD',
        'XT894',
        '_USB',
        'PROD_TAB13-201',
        'URFPAD2',
        'MID1126',
        'ST10216-1',
        'S5360L_CARD',
        'IDEATAB_A1000-F',
        'LBOOX',
        'LTAGUS',
        'IMCOV6L',
        '_101',
        'LPAPYRE_624',
        'S.L.',
    ]
    WINDOWS_CARD_A_MEM = [
        'ANDROID_PHONE',
        'GT-I9000_CARD',
        'SGH-I897',
        'FILE-STOR_GADGET',
        'SGH-T959_CARD',
        'SGH-T959',
        'SAMSUNG_ANDROID',
        'GT-P1000_CARD',
        'A70S',
        'A101IT',
        '7',
        'INCREDIBLE',
        'A7EB',
        'SGH-T849_CARD',
        '__UMS_COMPOSITE',
        'SGH-I997_CARD',
        'MB870',
        'ALPANDIGITAL',
        'ANDROID_MID',
        'P990_SD_CARD',
        '.K080',
        'LTE_CARD',
        'MB853',
        'A1-07___C0541A4F',
        'XT912',
        'MB855',
        'XT910',
        'BOOK_A10_CARD',
        'USB_2.0_DRIVER',
        'I9100T',
        'P999DW_SD_CARD',
        'KTABLET_PC',
        'FILE-CD_GADGET',
        'GT-I9001_CARD',
        'USB_2.0',
        'XT875',
        'UMS_COMPOSITE',
        'PRO',
        '.KOBO_VOX',
        'SGH-T989_CARD',
        'SGH-I727',
        'USB_FLASH_DRIVER',
        'ANDROID',
        'MID7042',
        '7035',
        'VIEWPAD_7E',
        'NOVO7',
        'ADVANCED',
        'TABLET_PC',
        'F',
        'E400_SD_CARD',
        'ST80208-1',
        'XT894',
        '_USB',
        'PROD_TAB13-201',
        'URFPAD2',
        'MID1126',
        'ANDROID_PLATFORM',
        'ST10216-1',
        'LBOOX',
        'LTAGUS',
        'IMCOV6L',
    ]

    OSX_MAIN_MEM = 'Android Device Main Memory'

    MAIN_MEMORY_VOLUME_LABEL = 'Android Device Main Memory'

    SUPPORTS_SUB_DIRS = True

    def post_open_callback(self):
        opts = self.settings()
        opts = opts.extra_customization
        if not opts:
            opts = [self.EBOOK_DIR_MAIN, '']

        def strtolist(x):
            if isinstance(x, (str, bytes)):
                x = [y.strip() for y in x.split(',')]
            return x or []

        opts = [strtolist(x) for x in opts]
        self._android_main_ebook_dir = opts[0]
        self._android_card_ebook_dir = opts[1]

    def get_main_ebook_dir(self, for_upload=False):
        dirs = self._android_main_ebook_dir
        if not for_upload:

            def aldiko_tweak(x):
                return 'eBooks' if x == 'eBooks/import' else x

            dirs = list(map(aldiko_tweak, dirs))
        return dirs

    def get_carda_ebook_dir(self, for_upload=False):
        if not for_upload:
            return ''
        return self._android_card_ebook_dir

    def get_cardb_ebook_dir(self, for_upload=False):
        return self.get_carda_ebook_dir()

    def windows_sort_drives(self, drives):
        try:
            assert self.device_being_opened is not None
            vid, pid, bcd = self.device_being_opened[:3]
        except Exception:
            vid, pid, bcd = -1, -1, -1
        if (vid, pid, bcd) == (0x0E79, 0x1408, 0x0222):
            letter_a = drives.get('carda', None)
            if letter_a is not None:
                drives['carda'] = drives['main']
                drives['main'] = letter_a
        return drives

    @classmethod
    def configure_for_kindle_app(cls):
        proxy = cls._configProxy()
        proxy['format_map'] = ['azw3', 'mobi', 'azw', 'azw1', 'azw4', 'pdf']
        proxy['use_subdirs'] = False
        proxy['extra_customization'] = [','.join(['kindle'] + cls.EBOOK_DIR_MAIN), '']

    @classmethod
    def configure_for_generic_epub_app(cls):
        proxy = cls._configProxy()
        del proxy['format_map']
        del proxy['use_subdirs']
        del proxy['extra_customization']


class S60(USBMS):
    name = 'S60 driver'
    gui_name = 'S60 phone'
    description = _('Communicate with S60 phones.')
    author = 'Kovid Goyal'
    supported_platforms = ['windows', 'osx', 'linux']

    VENDOR_ID = [0x421]
    PRODUCT_ID = [0x156]
    BCD = [0x100]

    # For use with zxreader
    FORMATS = ['fb2']
    EBOOK_DIR_MAIN = 'FB2 Books'

    VENDOR_NAME = 'NOKIA'
    WINDOWS_MAIN_MEM = 'S60'


class WEBOS(USBMS):
    name = 'WebOS driver'
    gui_name = 'WebOS Tablet'
    description = _('Communicate with WebOS tablets.')
    author = 'Kovid Goyal'
    supported_platforms = ['windows', 'osx', 'linux']

    # Ordered list of supported formats
    FORMATS = ['mobi', 'azw', 'prc']

    VENDOR_ID = [0x0830]
    PRODUCT_ID = [0x8074, 0x8072]
    BCD = [0x0327]

    EBOOK_DIR_MAIN = '.palmkindle'
    VENDOR_NAME = 'HP'
    WINDOWS_MAIN_MEM = 'WEBOS-DEVICE'

    THUMBNAIL_HEIGHT = 160
    THUMBNAIL_WIDTH = 120

    def upload_cover(self, path, filename, metadata, filepath):

        from PIL import Image, ImageDraw

        coverdata = getattr(metadata, 'thumbnail', None)
        if coverdata and coverdata[2]:
            cover = Image.open(io.BytesIO(coverdata[2]))
        else:
            coverdata = open(I('library.png'), 'rb').read()

            cover = Image.new('RGB', (120, 160), 'black')
            im = Image.open(io.BytesIO(coverdata))
            im.thumbnail((120, 160), Image.Resampling.LANCZOS)

            x, y = im.size
            cover.paste(im, ((120 - x) // 2, (160 - y) // 2))

            draw = ImageDraw.Draw(cover)
            draw.text((1, 10), metadata.get('title', _('Unknown')).encode('ascii', 'ignore'))
            draw.text((1, 140), metadata.get('authors', _('Unknown'))[0].encode('ascii', 'ignore'))

        data = io.BytesIO()
        cover.save(data, 'JPEG')
        coverdata = data.getvalue()

        with open(os.path.join(path, 'coverCache', filename + '-medium.jpg'), 'wb') as coverfile:
            coverfile.write(coverdata)
            fsync(coverfile)

        coverdata = getattr(metadata, 'thumbnail', None)
        if coverdata and coverdata[2]:
            cover = Image.open(io.BytesIO(coverdata[2]))
        else:
            coverdata = open(I('library.png'), 'rb').read()

            cover = Image.new('RGB', (52, 69), 'black')
            im = Image.open(io.BytesIO(coverdata))
            im.thumbnail((52, 69), Image.Resampling.LANCZOS)

            x, y = im.size
            cover.paste(im, ((52 - x) // 2, (69 - y) // 2))

        cover2 = cover.resize((52, 69), Image.Resampling.LANCZOS).convert('RGB')

        data = io.BytesIO()
        cover2.save(data, 'JPEG')
        coverdata = data.getvalue()

        with open(os.path.join(path, 'coverCache', filename + '-small.jpg'), 'wb') as coverfile:
            coverfile.write(coverdata)
            fsync(coverfile)



# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

import collections
import copy
import numbers
import operator
import re
import sys
from typing import cast

from qt.core import QBrush, QColor, QFont, QFontMetrics, QGraphicsItem, QGraphicsPixmapItem, QGraphicsRectItem, QPen, QPixmap, QRectF, Qt

from calibre.ebooks.hyphenate import hyphenate_word
from calibre.ebooks.lrf.fonts import LIBERATION_FONT_MAP


def WEIGHT_MAP(wt):
    return int(wt / 10 - 1)


def NULL(a, b):
    return a


def COLOR(a, b):
    return QColor(*a)


def WEIGHT(a, b):
    return WEIGHT_MAP(a)


class PixmapItem(QGraphicsPixmapItem):
    def __init__(self, data, encoding, x0, y0, x1, y1, xsize, ysize):
        p = QPixmap()
        p.loadFromData(data, encoding, Qt.ImageConversionFlag.AutoColor)
        w, h = p.width(), p.height()
        p = p.copy(x0, y0, min(w, x1 - x0), min(h, y1 - y0))
        if p.width() != xsize or p.height() != ysize:
            p = p.scaled(int(xsize), int(ysize), Qt.AspectRatioMode.IgnoreAspectRatio, Qt.TransformationMode.SmoothTransformation)
        QGraphicsPixmapItem.__init__(self, p)
        self.height, self.width = ysize, xsize
        self.setTransformationMode(Qt.TransformationMode.SmoothTransformation)
        self.setShapeMode(QGraphicsPixmapItem.ShapeMode.BoundingRectShape)

    def resize(self, width, height):
        p = self.pixmap()
        self.setPixmap(
            p.scaled(
                int(width),
                int(height),
                Qt.AspectRatioMode.IgnoreAspectRatio,
                Qt.TransformationMode.SmoothTransformation,
            )
        )
        self.width, self.height = width, height


class Plot(PixmapItem):
    def __init__(self, plot, dpi):
        img = plot.refobj
        xsize, ysize = dpi * plot.attrs['xsize'] / 720, dpi * plot.attrs['ysize'] / 720
        x0, y0, x1, y1 = img.x0, img.y0, img.x1, img.y1
        data, encoding = img.data, img.encoding
        PixmapItem.__init__(self, data, encoding, x0, y0, x1, y1, xsize, ysize)


class FontLoader:
    font_map = {
        'Swis721 BT Roman': 'Liberation Sans',
        'Dutch801 Rm BT Roman': 'Liberation Serif',
        'Courier10 BT Roman': 'Liberation Mono',
    }

    def __init__(self, font_map, dpi):
        self.face_map = {}
        self.cache = {}
        self.dpi = dpi
        self.face_map = font_map

    def font(self, text_style):
        device_font = text_style.fontfacename in LIBERATION_FONT_MAP
        try:
            if device_font:
                face = self.font_map[text_style.fontfacename]
            else:
                face = self.face_map[text_style.fontfacename]
        except KeyError:  # Bad fontfacename field in LRF
            face = self.font_map['Dutch801 Rm BT Roman']

        sz = text_style.fontsize
        wt = text_style.fontweight
        style = text_style.fontstyle
        font = (
            face,
            wt,
            style,
            sz,
        )
        if font in self.cache:
            rfont = self.cache[font]
        else:
            italic = font[2] == QFont.Style.StyleItalic
            rfont = QFont(font[0], int(font[3]), int(font[1]), italic)
            rfont.setPixelSize(int(font[3]))
            rfont.setBold(wt >= 69)
            self.cache[font] = rfont
        qfont = rfont
        if text_style.emplinetype != 'none':
            qfont = QFont(rfont)
            qfont.setOverline(text_style.emplineposition == 'before')
            qfont.setUnderline(text_style.emplineposition == 'after')
        return qfont


class Style:
    map = collections.defaultdict(lambda: NULL)

    def __init__(self, style, dpi):
        self.fdpi = dpi / 720
        self.update(style.as_dict())

    def update(self, *args, **kwds):
        if len(args) > 0:
            kwds = args[0]
        for attr in kwds:
            setattr(self, attr, self.__class__.map[attr](kwds[attr], self.fdpi))

    def copy(self):
        return copy.copy(self)


class TextStyle(Style):
    map = collections.defaultdict(
        lambda: NULL,
        fontsize=operator.mul,
        fontwidth=operator.mul,
        fontweight=WEIGHT,
        textcolor=COLOR,
        textbgcolor=COLOR,
        wordspace=operator.mul,
        letterspace=operator.mul,
        baselineskip=operator.mul,
        linespace=operator.mul,
        parindent=operator.mul,
        parskip=operator.mul,
        textlinewidth=operator.mul,
        charspace=operator.mul,
        linecolor=COLOR,
    )

    def __init__(self, style, font_loader, ruby_tags):
        self.font_loader = font_loader
        self.fontstyle = QFont.Style.StyleNormal
        for attr in ruby_tags:
            setattr(self, attr, ruby_tags[attr])
        Style.__init__(self, style, font_loader.dpi)
        self.emplinetype = 'none'
        self.font = self.font_loader.font(self)

    def update(self, *args, **kwds):
        Style.update(self, *args, **kwds)
        self.font = self.font_loader.font(self)


class BlockStyle(Style):
    map = collections.defaultdict(
        lambda: NULL,
        bgcolor=COLOR,
        framecolor=COLOR,
    )
    blockwidth: int
    blockheight: int
    sidemargin: int
    blockrule: str


class ParSkip:
    def __init__(self, parskip):
        self.height = parskip

    def __str__(self):
        return 'Parskip: ' + str(self.height)


class TextBlock:
    class HeightExceeded(Exception):
        pass

    has_content = property(fget=lambda self: self.peek_index < len(self.lines) - 1)
    XML_ENTITIES = {'apos': "'", 'quot': '"', 'amp': '&', 'lt': '<', 'gt': '>'}

    def __init__(self, tb, font_loader, respect_max_y, text_width, logger, opts, ruby_tags, link_activated):
        self.block_id = tb.id
        self.bs, self.ts = BlockStyle(tb.style, font_loader.dpi), TextStyle(tb.textstyle, font_loader, ruby_tags)
        self.bs.update(tb.attrs)
        self.ts.update(tb.attrs)
        self.lines = collections.deque()
        self.line_length = min(self.bs.blockwidth, text_width)
        self.line_length -= 2 * self.bs.sidemargin
        self.line_offset = self.bs.sidemargin
        self.first_line = True
        self.current_style = self.ts.copy()
        self.current_line = None
        self.font_loader, self.logger, self.opts = font_loader, logger, opts
        self.in_link = False
        self.link_activated = link_activated
        self.max_y = self.bs.blockheight if (respect_max_y or self.bs.blockrule.lower() in ('vert-fixed', 'block-fixed')) else sys.maxsize
        self.height = 0
        self.peek_index = -1

        try:
            self.populate(tb.content)
            self.end_line()
        except TextBlock.HeightExceeded:
            pass
            # logger.warning('TextBlock height exceeded, skipping line:\n%s'%(err,))

    def peek(self):
        return self.lines[self.peek_index + 1]

    def commit(self):
        self.peek_index += 1

    def reset(self):
        self.peek_index = -1

    def create_link(self, refobj):
        if self.current_line is None:
            self.create_line()
        assert self.current_line is not None
        self.current_line.start_link(refobj, self.link_activated)
        self.link_activated(refobj, on_creation=True)

    def end_link(self):
        if self.current_line is not None:
            self.current_line.end_link()

    def close_valign(self):
        if self.current_line is not None:
            self.current_line.valign = None

    def populate(self, tb):
        self.create_line()
        open_containers = collections.deque()
        self.in_para = False
        for i in tb.content:
            if isinstance(i, (str, bytes)):
                self.process_text(i)
            elif i is None:
                if len(open_containers) > 0:
                    for a, b in open_containers.pop():
                        if callable(a):
                            a(*b)
                        else:
                            setattr(self, a, b)
            elif i.name == 'P':
                open_containers.append((('in_para', False),))
                self.in_para = True
            elif i.name == 'CR':
                if self.in_para:
                    self.end_line()
                    self.create_line()
                else:
                    self.end_line()
                    delta = getattr(self.current_style, 'parskip', 0)
                    if isinstance(self.lines[-1], ParSkip):
                        delta += self.current_style.baselineskip
                    self.lines.append(ParSkip(delta))
                    self.first_line = True
            elif i.name == 'Span':
                open_containers.append((('current_style', self.current_style.copy()),))
                self.current_style.update(i.attrs)
            elif i.name == 'CharButton':
                open_containers.append(((self.end_link, []),))
                self.create_link(i.attrs['refobj'])
            elif i.name == 'Italic':
                open_containers.append((('current_style', self.current_style.copy()),))
                self.current_style.update(fontstyle=QFont.Style.StyleItalic)
            elif i.name == 'Plot':
                plot = Plot(i, self.font_loader.dpi)
                if self.current_line is None:
                    self.create_line()
                assert self.current_line is not None
                if not self.current_line.can_add_plot(plot):
                    self.end_line()
                    self.create_line()
                assert self.current_line is not None
                self.current_line.add_plot(plot)
            elif i.name in ['Sup', 'Sub']:
                if self.current_line is None:
                    self.create_line()
                assert self.current_line is not None
                self.current_line.valign = i.name
                open_containers.append(((self.close_valign, []),))
            elif i.name == 'Space' and self.current_line is not None:
                self.current_line.add_space(i.attrs['xsize'])
            elif i.name == 'EmpLine':
                if i.attrs:
                    open_containers.append((('current_style', self.current_style.copy()),))
                    self.current_style.update(i.attrs)
            else:
                self.logger.warning(f'Unhandled TextTag {i.name}')
                if not i.self_closing:
                    open_containers.append([])

    def end_line(self):
        if self.current_line is not None:
            self.height += self.current_line.finalize(self.current_style.baselineskip, self.current_style.linespace, self.opts.visual_debug)
            if self.height > self.max_y + 10:
                raise TextBlock.HeightExceeded(str(self.current_line))
            self.lines.append(self.current_line)
            self.current_line = None

    def create_line(self):
        line_length = self.line_length
        line_offset = self.line_offset
        if self.first_line:
            line_length -= self.current_style.parindent
            line_offset += self.current_style.parindent
        self.current_line = Line(
            line_length,
            line_offset,
            self.current_style.linespace,
            self.current_style.align,
            self.opts.hyphenate,
            self.block_id,
        )
        self.first_line = False

    def process_text(self, raw):
        for ent, rep in TextBlock.XML_ENTITIES.items():
            raw = raw.replace(f'&{ent};', rep)
        while len(raw) > 0:
            if self.current_line is None:
                self.create_line()
            assert self.current_line is not None
            pos, line_filled = self.current_line.populate(raw, self.current_style)
            raw = raw[pos:]
            if line_filled:
                self.end_line()
            if not pos:
                break

    def __iter__(self):
        yield from self.lines

    def __str__(self):
        s = ''
        for line in self:
            s += str(line) + '\n'
        return s


class Link(QGraphicsRectItem):
    inactive_brush = QBrush(QColor(0xFF, 0xFF, 0xFF, 0xFF))
    active_brush = QBrush(QColor(0x00, 0x00, 0x00, 0x59))

    def __init__(self, parent, start, stop, refobj, slot):
        QGraphicsRectItem.__init__(self, start, 0, stop - start, parent.height, parent)
        self.refobj = refobj
        self.slot = slot
        self._brush = self.__class__.inactive_brush
        self.setPen(QPen(Qt.PenStyle.NoPen))
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setAcceptHoverEvents(True)

    def hoverEnterEvent(self, event):
        self._brush = self.__class__.active_brush
        parent_item = self.parentItem()
        assert parent_item is not None
        parent_item.update()

    def hoverLeaveEvent(self, event):
        self._brush = self.__class__.inactive_brush
        parent_item = self.parentItem()
        assert parent_item is not None
        parent_item.update()

    def mousePressEvent(self, event):
        self.hoverLeaveEvent(None)
        self.slot(self.refobj)


class Line(QGraphicsItem):
    whitespace = re.compile(r'\s+')

    def __init__(self, line_length, offset, linespace, align, hyphenate, block_id):
        QGraphicsItem.__init__(self)

        self.line_length, self.offset, self.line_space = line_length, offset, linespace
        self.align, self.hyphenate, self.block_id = align, hyphenate, block_id

        self.tokens = collections.deque()
        self.current_width = 0
        self.length_in_space = 0
        self.height, self.descent, self.width = 0, 0, 0
        self.links = collections.deque()
        self.current_link = None
        self.valign: str | None = None

    def start_link(self, refobj, slot):
        self.current_link = [self.current_width, sys.maxsize, refobj, slot]

    def end_link(self):
        if self.current_link is not None:
            self.current_link[1] = self.current_width
            self.links.append(self.current_link)
            self.current_link = None

    def can_add_plot(self, plot):
        return self.line_length - self.current_width >= plot.width

    def add_plot(self, plot):
        self.tokens.append(plot)
        self.current_width += plot.width
        self.height = max(self.height, plot.height)
        self.add_space(6)

    def populate(self, phrase, ts, process_space=True):
        phrase_pos = 0
        processed = False
        matches = self.__class__.whitespace.finditer(phrase)
        font = QFont(ts.font)
        if self.valign is not None:
            font.setPixelSize(int(font.pixelSize() / 1.5))
        fm = QFontMetrics(font)
        single_space_width = fm.horizontalAdvance(' ')
        height, descent = fm.height(), fm.descent()
        for match in matches:
            processed = True
            left, right = match.span()
            if not process_space:
                right = left
            space_width = single_space_width * (right - left)
            word = phrase[phrase_pos:left]
            width = fm.horizontalAdvance(word)
            if self.current_width + width < self.line_length:
                self.commit(word, width, height, descent, ts, font)
                if space_width > 0 and self.current_width + space_width < self.line_length:
                    self.add_space(space_width)
                phrase_pos = right
                continue

            # Word doesn't fit on line
            if self.hyphenate and len(word) > 3:
                tokens = hyphenate_word(word)
                for i in range(len(tokens) - 2, -1, -1):
                    word = ''.join(tokens[0 : i + 1]) + '-'
                    width = fm.horizontalAdvance(word)
                    if self.current_width + width < self.line_length:
                        self.commit(word, width, height, descent, ts, font)
                        return phrase_pos + len(word) - 1, True
            if self.current_width < 5:  # Force hyphenation as word is longer than line
                for i in range(len(word) - 5, 0, -5):
                    part = word[:i] + '-'
                    width = fm.horizontalAdvance(part)
                    if self.current_width + width < self.line_length:
                        self.commit(part, width, height, descent, ts, font)
                        return phrase_pos + len(part) - 1, True
            # Failed to add word.
            return phrase_pos, True

        if not processed:
            return self.populate(phrase + ' ', ts, False)

        return phrase_pos, False

    def commit(self, word, width, height, descent, ts, font):
        self.tokens.append(Word(word, width, height, ts, font, self.valign))
        self.current_width += width
        self.height = max(self.height, height)
        self.descent = max(self.descent, descent)

    def add_space(self, min_width):
        self.tokens.append(min_width)
        self.current_width += min_width
        self.length_in_space += min_width

    def justify(self):
        delta = self.line_length - self.current_width
        if self.length_in_space > 0:
            frac = 1 + float(delta) / self.length_in_space
            for i in range(len(self.tokens)):
                if isinstance(self.tokens[i], numbers.Number):
                    self.tokens[i] *= frac
            self.current_width = self.line_length

    def finalize(self, baselineskip, linespace, vdebug):
        if self.current_link is not None:
            self.end_link()

        # We justify if line is small and it doesn't have links in it
        # If it has links, justification would cause the boundingrect of the link to
        # be too small
        if self.current_width >= 0.85 * self.line_length and len(self.links) == 0:
            self.justify()

        self.width = float(self.current_width)
        if self.height == 0:
            self.height = baselineskip
        self.height = float(self.height)

        self.vdebug = vdebug

        for link in self.links:
            Link(self, *link)

        return self.height

    def boundingRect(self):
        return QRectF(0, 0, self.width, self.height)

    def paint(self, painter, option, widget=None):
        x, y = 0, 0 + self.height - self.descent
        if self.vdebug:
            painter.save()
            painter.setPen(QPen(Qt.GlobalColor.yellow, 1, Qt.PenStyle.DotLine))
            painter.drawRect(self.boundingRect())
            painter.restore()
        painter.save()
        painter.setPen(QPen(Qt.PenStyle.NoPen))
        for c in self.childItems():
            painter.setBrush(cast('Link', c)._brush)
            painter.drawRect(c.boundingRect())
        painter.restore()
        painter.save()
        for tok in self.tokens:
            if isinstance(tok, numbers.Number):
                x += tok
            elif isinstance(tok, Word):
                painter.setFont(tok.font)
                if tok.highlight:
                    painter.save()
                    painter.setPen(QPen(Qt.PenStyle.NoPen))
                    painter.setBrush(QBrush(Qt.GlobalColor.yellow))
                    painter.drawRect(int(x), 0, tok.width, tok.height)
                    painter.restore()
                painter.setPen(QPen(tok.text_color))
                if tok.valign is None:
                    painter.drawText(int(x), int(y), tok.string)
                elif tok.valign == 'Sub':
                    painter.drawText(int(x + 1), int(y + self.descent / 1.5), tok.string)
                elif tok.valign == 'Sup':
                    painter.drawText(int(x + 1), int(y - 2.0 * self.descent), tok.string)
                x += tok.width
            else:
                painter.drawPixmap(int(x), 0, tok.pixmap())
                x += tok.width
        painter.restore()

    def words(self):
        for w in self.tokens:
            if isinstance(w, Word):
                yield w

    def search(self, phrase):
        tokens = phrase.lower().split()
        if len(tokens) < 1:
            return None

        words = self.words()
        matches = []
        try:
            while True:
                word = next(words)
                word.highlight = False
                if tokens[0] in str(word.string).lower():
                    matches.append(word)
                    for c in range(1, len(tokens)):
                        word = next(words)
                        print(tokens[c], word.string)
                        if tokens[c] not in str(word.string):
                            return None
                        matches.append(word)
                    for w in matches:
                        w.highlight = True
                    return self
        except StopIteration:
            return None

    def getx(self, textwidth):
        if self.align == 'head':
            return self.offset
        if self.align == 'foot':
            return textwidth - self.width
        if self.align == 'center':
            return (textwidth - self.width) / 2.0

    def __unicode__(self):
        s = ''
        for tok in self.tokens:
            if isinstance(tok, numbers.Number):
                s += ' '
            elif isinstance(tok, Word):
                s += str(tok.string)
        return s

    def __str__(self):
        return str(self).encode('utf-8')


class Word:
    def __init__(self, string, width, height, ts, font, valign):
        self.string, self.width, self.height = string, width, height
        self.font = font
        self.text_color = ts.textcolor
        self.highlight = False
        self.valign = valign


def main(args=sys.argv):
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# License: GPLv3 Copyright: 2011, John Schember <john@nachtimwald.com>

from __future__ import absolute_import, division, print_function, unicode_literals

store_version = 4  # Needed for dynamic plugin loading

from contextlib import closing
from urllib.parse import quote

from qt.core import QUrl

from calibre import browser, url_slash_cleaner
from calibre.gui2 import open_url
from calibre.gui2.store import StorePlugin
from calibre.gui2.store.basic_config import BasicStoreConfig
from calibre.gui2.store.search_result import SearchResult
from calibre.gui2.store.web_store_dialog import WebStoreDialog

try:
    from calibre.utils.xml_parse import safe_html_fromstring
except ImportError:
    from lxml.html import fromstring as safe_html_fromstring


class MillsBoonUKStore(BasicStoreConfig, StorePlugin):
    def open(self, gui=None, parent=None, detail_item=None, external=False):
        url = 'https://www.millsandboon.co.uk'

        if external or self.config.get('open_external', False):
            if detail_item:
                url = detail_item
            open_url(QUrl(url_slash_cleaner(url)))
        else:
            if detail_item:
                detail_url = detail_item
            else:
                detail_url = None
            d = WebStoreDialog(self.gui, url, parent, detail_url)
            d.setWindowTitle(self.name)
            d.set_tags(self.config.get('tags', ''))
            d.exec()

    def search(self, query, max_results=10, timeout=60):
        base_url = 'https://www.millsandboon.co.uk'
        url = base_url + '/search.aspx??format=ebook&searchText=' + quote(query)
        br = browser()

        counter = max_results
        with closing(br.open(url, timeout=timeout)) as f:
            doc = safe_html_fromstring(f.read())
            for data in doc.xpath('//article[contains(@class, "group")]'):
                if counter <= 0:
                    break
                id_ = ''.join(data.xpath('.//div[@class="img-wrapper"]/a/@href')).strip()
                if not id_:
                    continue

                cover_url = ''.join(data.xpath('.//div[@class="img-wrapper"]/a/img/@src'))
                title = ''.join(data.xpath('.//div[@class="img-wrapper"]/a/img/@alt')).strip()
                author = ''.join(data.xpath('.//a[@class="author"]/text()'))
                price = ''.join(data.xpath('.//div[@class="type-wrapper"]/ul/li[child::span[text()="eBook"]]/a/text()'))
                format_ = ''.join(data.xpath('.//p[@class="doc-meta-format"]/span[last()]/text()'))
                drm = SearchResult.DRM_LOCKED

                counter -= 1

                s = SearchResult()
                s.cover_url = cover_url
                s.title = title.strip()
                s.author = author.strip()
                s.price = price
                s.detail_item = id_
                s.drm = drm
                s.formats = format_

                yield s



# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

import re

NAME_MAP = {
    'aliceblue': '#F0F8FF',
    'antiquewhite': '#FAEBD7',
    'aqua': '#00FFFF',
    'aquamarine': '#7FFFD4',
    'azure': '#F0FFFF',
    'beige': '#F5F5DC',
    'bisque': '#FFE4C4',
    'black': '#000000',
    'blanchedalmond': '#FFEBCD',
    'blue': '#0000FF',
    'brown': '#A52A2A',
    'burlywood': '#DEB887',
    'cadetblue': '#5F9EA0',
    'chartreuse': '#7FFF00',
    'chocolate': '#D2691E',
    'coral': '#FF7F50',
    'crimson': '#DC143C',
    'cyan': '#00FFFF',
    'darkblue': '#00008B',
    'darkgoldenrod': '#B8860B',
    'darkgreen': '#006400',
    'darkkhaki': '#BDB76B',
    'darkmagenta': '#8B008B',
    'darkolivegreen': '#556B2F',
    'darkorange': '#FF8C00',
    'darkorchid': '#9932CC',
    'darkred': '#8B0000',
    'darksalmon': '#E9967A',
    'darkslateblue': '#483D8B',
    'darkslategrey': '#2F4F4F',
    'darkviolet': '#9400D3',
    'deeppink': '#FF1493',
    'dodgerblue': '#1E90FF',
    'firebrick': '#B22222',
    'floralwhite': '#FFFAF0',
    'forestgreen': '#228B22',
    'fuchsia': '#FF00FF',
    'gainsboro': '#DCDCDC',
    'ghostwhite': '#F8F8FF',
    'gold': '#FFD700',
    'goldenrod': '#DAA520',
    'indianred ': '#CD5C5C',
    'indigo  ': '#4B0082',
    'khaki': '#F0E68C',
    'lavenderblush': '#FFF0F5',
    'lawngreen': '#7CFC00',
    'lightblue': '#ADD8E6',
    'lightcoral': '#F08080',
    'lightgoldenrodyellow': '#FAFAD2',
    'lightgray': '#D3D3D3',
    'lightgrey': '#D3D3D3',
    'lightskyblue': '#87CEFA',
    'lightslategrey': '#778899',
    'lightsteelblue': '#B0C4DE',
    'lime': '#87CEFA',
    'linen': '#FAF0E6',
    'magenta': '#FF00FF',
    'maroon': '#800000',
    'mediumaquamarine': '#66CDAA',
    'mediumblue': '#0000CD',
    'mediumorchid': '#BA55D3',
    'mediumpurple': '#9370D8',
    'mediumseagreen': '#3CB371',
    'mediumslateblue': '#7B68EE',
    'midnightblue': '#191970',
    'moccasin': '#FFE4B5',
    'navajowhite': '#FFDEAD',
    'navy': '#000080',
    'oldlace': '#FDF5E6',
    'olive': '#808000',
    'orange': '#FFA500',
    'orangered': '#FF4500',
    'orchid': '#DA70D6',
    'paleturquoise': '#AFEEEE',
    'papayawhip': '#FFEFD5',
    'peachpuff': '#FFDAB9',
    'powderblue': '#B0E0E6',
    'rosybrown': '#BC8F8F',
    'royalblue': '#4169E1',
    'saddlebrown': '#8B4513',
    'sandybrown': '#8B4513',
    'seashell': '#FFF5EE',
    'sienna': '#A0522D',
    'silver': '#C0C0C0',
    'skyblue': '#87CEEB',
    'slategrey': '#708090',
    'snow': '#FFFAFA',
    'springgreen': '#00FF7F',
    'violet': '#EE82EE',
    'yellowgreen': '#9ACD32',
}

hex_pat = re.compile(r'#(\d{2})(\d{2})(\d{2})')
rgb_pat = re.compile(r'rgb\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)', re.IGNORECASE)


def lrs_color(html_color):
    hcol = html_color.lower()
    match = hex_pat.search(hcol)
    if match:
        return '0x00' + match.group(1) + match.group(2) + match.group(3)
    match = rgb_pat.search(hcol)
    if match:
        return f'0x00{int(match.group(1)):x}{int(match.group(2)):x}{int(match.group(3)):x}'
    if hcol in NAME_MAP:
        return NAME_MAP[hcol].replace('#', '0x00')
    return '0x00000000'



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

import json
import sys
import time
import weakref
from collections import defaultdict, deque
from datetime import datetime
from itertools import chain
from operator import attrgetter

from calibre import force_unicode, human_readable, prints
from calibre.constants import iswindows
from calibre.ebooks import BOOK_EXTENSIONS
from calibre.utils.date import as_utc, local_tz
from calibre.utils.icu import lower, sort_key

bexts = frozenset(BOOK_EXTENSIONS) - {'mbp', 'tan', 'rar', 'zip', 'xml'}


def convert_timestamp(md):
    try:
        if isinstance(md, tuple):
            year, month, day, hour, minute, second, microsecond = md
            return datetime(year, month, day, hour, minute, second, microsecond, local_tz)
        else:
            return datetime.fromtimestamp(md, local_tz)
    except Exception:
        return datetime.fromtimestamp(0, local_tz)


class ListEntry:
    def __init__(self, entry: FileOrFolder):
        self.is_dir = entry.is_folder
        self.is_readonly = not entry.can_delete
        self.path = '/'.join(entry.full_path)
        self.name = entry.name
        self.size = entry.size
        self.ctime = self.wtime = time.mktime(entry.last_modified.timetuple())


class FileOrFolder:
    def __init__(self, entry, fs_cache: FilesystemCache, is_storage: bool = False):
        self.object_id = entry['id']
        self.is_storage = is_storage
        self.is_folder = entry['is_folder']
        self.storage_id = entry['storage_id']
        # self.parent_id is None for storage objects
        self.parent_id = entry.get('parent_id', None)
        self.persistent_id = entry.get('persistent_id', self.object_id)
        n = entry.get('name', None)
        if not n:
            if self.is_storage:
                prefix = 'Storage'
            else:
                prefix = 'Folder' if self.is_folder else 'File'
            n = f'{prefix}-{self.persistent_id}'
        self.name = force_unicode(n, 'utf-8')
        self.size = entry.get('size', 0)
        md = entry.get('modified', 0)
        self.last_modified = convert_timestamp(md)
        self.last_mod_string = self.last_modified.strftime('%Y/%m/%d %H:%M')
        self.last_modified = as_utc(self.last_modified)

        if self.storage_id not in fs_cache.all_storage_ids:
            raise ValueError(f'Storage id {self.storage_id} not valid for {entry}, valid values: {fs_cache.all_storage_ids}')

        self.is_hidden = entry.get('is_hidden', False)
        self.is_system = entry.get('is_system', False)
        self.can_delete = entry.get('can_delete', True)

        self.files = []
        self.folders = []
        if not self.is_storage:
            # storage ids can overlap filesystem object ids on libmtp. See https://bugs.launchpad.net/bugs/2072384
            # so only store actual filesystem object ids in id_map
            fs_cache.id_maps[self.storage_id][self.object_id] = self
            if iswindows:
                # On windows parent_id == storage_id for objects in root. Set
                # it 0 so the rest of the logic works as on libmtp.
                # See https://bugs.launchpad.net/bugs/2073323
                if self.storage_id == self.parent_id:
                    self.parent_id = 0
        self.fs_cache = weakref.ref(fs_cache)
        self.deleted = False

        if self.is_storage:
            self.storage_prefix = f'mtp:::{self.persistent_id}:::'

        # Ignore non ebook files and AppleDouble files
        self.is_ebook = not self.is_folder and not self.is_storage and self.name.rpartition('.')[-1].lower() in bexts and not self.name.startswith('._')

        # allow Kindle Scribe notebooks to be imported and preserve the notebook name
        if self.name == 'nbk' and not (self.is_folder or self.is_storage):
            if len(self.full_path) >= 3 and self.full_path[-3] == '.notebooks':
                nbk_name = self.full_path[-2].replace('!!notebook', '').replace('!!', ' ')
                self.name = f'Scribe {nbk_name} Notebook.kfx'
                self.is_ebook = True

    def __repr__(self):
        if self.is_storage:
            name = 'Storage'
        else:
            name = 'Folder' if self.is_folder else 'File'
        try:
            path = str(self.full_path)
        except Exception:
            path = ''
        datum = f'size={self.size}'
        if self.is_folder or self.is_storage:
            datum = f'children={len(self.files) + len(self.folders)}'
        return f'{name}(id={self.object_id}, storage_id={self.storage_id}, {datum}, path={path}, modified={self.last_mod_string})'

    __str__ = __repr__
    __unicode__ = __repr__

    @property
    def empty(self):
        return not self.files and not self.folders

    @property
    def id_map(self) -> dict[int, FileOrFolder]:
        fs_cache = self.fs_cache()
        assert fs_cache is not None
        return fs_cache.id_maps[self.storage_id]

    @property
    def parent(self):
        if self.parent_id:
            return self.id_map[self.parent_id]
        if self.is_storage or self.parent_id is None:
            return None
        fs_cache2 = self.fs_cache()
        assert fs_cache2 is not None
        return fs_cache2.storage(self.storage_id)

    @property
    def in_root(self):
        return self.parent_id is not None and self.parent_id == 0

    @property
    def storage(self):
        fs_cache3 = self.fs_cache()
        assert fs_cache3 is not None
        return fs_cache3.storage(self.storage_id)

    @property
    def full_path(self) -> tuple[str, ...]:
        parts = deque()
        parts.append(self.name)
        p = self.parent
        while p is not None:
            parts.appendleft(p.name)
            p = p.parent
        return tuple(parts)

    def __iter__(self):
        yield from self.folders
        yield from self.files

    def add_child(self, entry):
        cache = self.fs_cache()
        assert cache is not None
        ans = FileOrFolder(entry, cache)
        t = self.folders if ans.is_folder else self.files
        t.append(ans)
        return ans

    def remove_child(self, entry):
        for x in (self.files, self.folders):
            try:
                x.remove(entry)
            except ValueError:
                pass
        self.id_map.pop(entry.object_id, None)
        entry.deleted = True

    def dump(self, prefix='', out=sys.stdout):
        c = '+' if self.is_folder else '-'
        data = f'{sum(map(len, (self.files, self.folders)))} children' if self.is_folder else human_readable(self.size)
        data += f' modified={self.last_mod_string}'
        line = f'{prefix}{c} {self.name} [id:{self.object_id} {data}]'
        prints(line, file=out)
        for c in (self.folders, self.files):
            for e in sorted(c, key=lambda x: sort_key(x.name)):
                e.dump(prefix=prefix + '  ', out=out)

    def list(self, recurse=False):
        if not self.is_folder:
            parent = self.parent
            yield '/'.join(parent.full_path[1:]), ListEntry(self)
            return
        entries = [ListEntry(x) for x in chain(self.folders, self.files)]
        yield '/'.join(self.full_path[1:]), entries
        if recurse:
            for x in self.folders:
                yield from x.list(recurse=True)

    def folder_named(self, name):
        name = lower(name)
        for e in self.folders:
            if e.name and lower(e.name) == name:
                return e
        return None

    def file_named(self, name):
        name = lower(name)
        for e in self.files:
            if e.name and lower(e.name) == name:
                return e
        return None

    def find_path(self, path):
        """
        Find a path in this folder, where path is a
        tuple of folder and file names like ('eBooks', 'newest',
        'calibre.epub'). Finding is case-insensitive.
        """
        parent = self
        components = list(path)
        while components:
            child = components[0]
            components = components[1:]
            c = parent.folder_named(child)
            if c is None:
                c = parent.file_named(child)
            if c is None:
                return None
            parent = c
        return parent

    @property
    def mtp_relpath(self):
        return tuple(x.lower() for x in self.full_path[1:])

    @property
    def mtp_id_path(self):
        return 'mtp:::' + json.dumps(self.object_id) + ':::' + '/'.join(self.full_path)


class FilesystemCache:
    def __init__(self, all_storage, entries):
        self.entries = []
        self.id_maps = defaultdict(dict)
        self.all_storage_ids = tuple(x['id'] for x in all_storage)

        for storage in all_storage:
            storage['storage_id'] = storage['id']
            e = FileOrFolder(storage, self, is_storage=True)
            self.entries.append(e)

        self.entries.sort(key=attrgetter('object_id'))
        self.all_storage_ids = tuple(x.storage_id for x in self.entries)

        for entry in entries:
            FileOrFolder(entry, self)

        for id_map in self.id_maps.values():
            for item in id_map.values():
                try:
                    p = item.parent
                except KeyError:
                    # Parent does not exist, set the parent to be the storage
                    # object
                    item.parent_id = 0
                    p = item.parent

                if p is not None:
                    t = p.folders if item.is_folder else p.files
                    t.append(item)

    def dump(self, out=sys.stdout):
        for e in self.entries:
            e.dump(out=out)

    def storage(self, storage_id):
        for e in self.entries:
            if e.storage_id == storage_id:
                return e

    def iterebooks(self, storage_id):
        id_map = self.id_maps[storage_id]
        for x in id_map.values():
            if x.is_ebook:
                if x.in_root and x.name.lower().endswith('.txt'):
                    continue  # Ignore .txt files in the root
                yield x

    def __len__(self):
        ans = len(self.id_maps)
        for id_map in self.id_maps.values():
            ans += len(id_map)
        return ans

    def resolve_mtp_id_path(self, path):
        if not path.startswith('mtp:::'):
            raise ValueError(f'{path} is not a valid MTP path')
        parts = path.split(':::', 2)
        if len(parts) < 3:
            raise ValueError(f'{path} is not a valid MTP path')
        try:
            object_id = json.loads(parts[1])
        except Exception:
            raise ValueError(f'{path} is not a valid MTP path')
        id_map = {}
        path = parts[2]
        storage_name = path.partition('/')[0]
        for entry in self.entries:
            if entry.name == storage_name:
                id_map = self.id_maps[entry.storage_id]
                break
        try:
            return id_map[object_id]
        except KeyError:
            raise ValueError(f'No object found with MTP path: {path}')



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid@kovidgoyal.net>

import struct
from collections import OrderedDict, namedtuple

from calibre.ebooks.mobi.reader.headers import NULL_INDEX
from calibre.ebooks.mobi.reader.index import CNCX, INDEX_HEADER_FIELDS, get_tag_section_start, parse_index_record, parse_indx_header, parse_tagx_section
from calibre.ebooks.mobi.reader.ncx import default_entry, tag_fieldname_map

File = namedtuple('File', 'file_number name divtbl_count start_position length')

Elem = namedtuple('Elem', 'insert_pos toc_text file_number sequence_number start_pos length')

GuideRef = namedtuple('GuideRef', 'type title pos_fid')

INDEX_HEADER_FIELDS = INDEX_HEADER_FIELDS + ('indices', 'tagx_block_size', 'tagx_block')
FIELD_NAMES = {
    'len': 'Header length',
    'type': 'Unknown',
    'gen': 'Index Type (0 - normal, 2 - inflection)',
    'start': 'IDXT Offset',
    'count': 'Number of entries in this record',
    'code': 'character encoding',
    'lng': 'Unknown',
    'total': 'Total number of actual Index Entries in all records',
    'ordt': 'ORDT Offset',
    'ligt': 'LIGT Offset',
    'nligt': 'Number of LIGT',
    'ncncx': 'Number of CNCX records',
    'indices': 'Geometry of index records',
}


def read_variable_len_data(data, header):
    offset = header['tagx']
    indices = []
    idxt_offset = header['start']
    idxt_size = 4 + header['count'] * 2
    if offset > 0:
        tagx_block_size = header['tagx_block_size'] = struct.unpack_from(b'>I', data, offset + 4)[0]
        header['tagx_block'] = data[offset : offset + tagx_block_size]
        offset = idxt_offset + 4
        for i in range(header['count']):
            p = struct.unpack_from(b'>H', data, offset)[0]
            offset += 2
            strlen = bytearray(data[p])[0]
            text = data[p + 1 : p + 1 + strlen]
            p += 1 + strlen
            num = struct.unpack_from(b'>H', data, p)[0]
            indices.append((text, num))
    else:
        header['tagx_block'] = b''
        header['tagx_block_size'] = 0
    trailing_bytes = data[idxt_offset + idxt_size :]
    if trailing_bytes.rstrip(b'\0'):
        raise ValueError('Trailing bytes after last IDXT entry: {!r}'.format(trailing_bytes.rstrip(b'\0')))
    header['indices'] = indices


def read_index(sections, idx, codec):
    table, cncx = OrderedDict(), CNCX([], codec)

    data = sections[idx].raw

    indx_header = parse_indx_header(data)
    indx_count = indx_header['count']

    if indx_header['ncncx'] > 0:
        off = idx + indx_count + 1
        cncx_records = [x.raw for x in sections[off : off + indx_header['ncncx']]]
        cncx = CNCX(cncx_records, codec)

    tag_section_start = get_tag_section_start(data, indx_header)
    control_byte_count, tags = parse_tagx_section(data[tag_section_start:])

    read_variable_len_data(data, indx_header)
    index_headers = []

    for i in range(idx + 1, idx + 1 + indx_count):
        # Index record
        data = sections[i].raw
        index_headers.append(parse_index_record(table, data, control_byte_count, tags, codec, indx_header['ordt_map'], strict=True))
        read_variable_len_data(data, index_headers[-1])
    return table, cncx, indx_header, index_headers


class Index:
    def __init__(self, idx, records, codec):
        self.table = self.cncx = self.header = None
        self.records: list = []
        self.index_headers = []
        if idx != NULL_INDEX:
            self.table, self.cncx, self.header, self.index_headers = read_index(records, idx, codec)

    def render(self):
        ans = ['*' * 10 + ' Index Header ' + '*' * 10]
        a = ans.append
        if self.header is not None:
            for field in INDEX_HEADER_FIELDS:
                a(f'{FIELD_NAMES.get(field, field):<12}: {self.header[field]!r}')
        ans.extend(['', ''])
        ans += ['*' * 10 + f' Index Record Headers ({len(self.index_headers)} records) ' + '*' * 10]
        for i, header in enumerate(self.index_headers):
            ans += ['*' * 10 + f' Index Record {i} ' + '*' * 10]
            for field in INDEX_HEADER_FIELDS:
                a(f'{FIELD_NAMES.get(field, field):<12}: {header[field]!r}')

        if self.cncx:
            a('*' * 10 + ' CNCX ' + '*' * 10)
            for offset, val in self.cncx.items():
                a(f'{offset:10}: {val}')
            ans.extend(['', ''])

        if self.table is not None:
            a('*' * 10 + f' {len(self.table)} Index Entries ' + '*' * 10)
            for k, v in self.table.items():
                a(f'{k}: {v!r}')

        if self.records:
            ans.extend(['', '', '*' * 10 + ' Parsed Entries ' + '*' * 10])
            for f in self.records:
                a(repr(f))

        return ans + ['']

    def __str__(self):
        return '\n'.join(self.render())

    def __iter__(self):
        return iter(self.records)


class SKELIndex(Index):
    def __init__(self, skelidx, records, codec):
        super().__init__(skelidx, records, codec)
        self.records = []

        if self.table is not None:
            for i, text in enumerate(self.table):
                tag_map = self.table[text]
                if set(tag_map) != {1, 6}:
                    raise ValueError(f'SKEL Index has unknown tags: {set(tag_map) - {1, 6}}')
                self.records.append(
                    File(
                        i,  # file_number
                        text,  # name
                        tag_map[1][0],  # divtbl_count
                        tag_map[6][0],  # start_pos
                        tag_map[6][1],
                    )  # length
                )


class SECTIndex(Index):
    def __init__(self, sectidx, records, codec):
        super().__init__(sectidx, records, codec)
        self.records = []

        if self.table is not None:
            assert self.cncx is not None
            for i, text in enumerate(self.table):
                tag_map = self.table[text]
                if set(tag_map) != {2, 3, 4, 6}:
                    raise ValueError(f'Chunk Index has unknown tags: {set(tag_map) - {2, 3, 4, 6}}')

                toc_text = self.cncx[tag_map[2][0]]
                self.records.append(
                    Elem(
                        int(text),  # insert_pos
                        toc_text,  # toc_text
                        tag_map[3][0],  # file_number
                        tag_map[4][0],  # sequence_number
                        tag_map[6][0],  # start_pos
                        tag_map[6][1],  # length
                    )
                )


class GuideIndex(Index):
    def __init__(self, guideidx, records, codec):
        super().__init__(guideidx, records, codec)
        self.records = []

        if self.table is not None:
            assert self.cncx is not None
            for i, text in enumerate(self.table):
                tag_map = self.table[text]
                if set(tag_map) not in ({1, 6}, {1, 2, 3}):
                    raise ValueError(f'Guide Index has unknown tags: {tag_map}')

                title = self.cncx[tag_map[1][0]]
                self.records.append(GuideRef(text, title, tag_map[6] if 6 in tag_map else (tag_map[2], tag_map[3])))


class NCXIndex(Index):
    def __init__(self, ncxidx, records, codec):
        super().__init__(ncxidx, records, codec)
        self.records = []

        if self.table is not None:
            NCXEntry = namedtuple('NCXEntry', 'index start length depth parent first_child last_child title pos_fid kind')

            for num, x in enumerate(self.table.items()):
                text, tag_map = x
                e: dict = default_entry.copy()
                e['name'] = text
                e['num'] = num

                for tag in tag_fieldname_map:
                    fieldname, i = tag_fieldname_map[tag]
                    assert isinstance(fieldname, str)
                    if tag in tag_map:
                        fieldvalue = tag_map[tag][i]
                        if tag == 6:
                            # Appears to be an idx into the KF8 elems table with an
                            # offset
                            fieldvalue = tuple(tag_map[tag])
                        e[fieldname] = fieldvalue
                        for which, name in {
                            3: 'text',
                            5: 'kind',
                            70: 'description',
                            71: 'author',
                            72: 'image_caption',
                            73: 'image_attribution',
                        }.items():
                            if tag == which:
                                cncx = self.cncx
                                assert cncx is not None
                                e[name] = cncx.get(fieldvalue, default_entry[name])

                def refindx(e, name):
                    ans = e[name]
                    if ans < 0:
                        ans = None
                    return ans

                entry = NCXEntry(
                    start=e['pos'],
                    index=e['num'],
                    length=e['len'],
                    depth=e['hlvl'],
                    parent=refindx(e, 'parent'),
                    first_child=refindx(e, 'child1'),
                    last_child=refindx(e, 'childn'),
                    title=e['text'],
                    pos_fid=e['pos_fid'],
                    kind=e['kind'],
                )
                self.records.append(entry)
